#!/usr/bin/env python
"""
usage: subaligner_train [-h] -tod TRAINING_OUTPUT_DIRECTORY [-vd VIDEO_DIRECTORY] [-sd SUBTITLE_DIRECTORY] [-r] [-dde] [-bs BATCH_SIZE] [-do DROPOUT] [-e EPOCHS] [-p PATIENCE]
                        [-fhs FRONT_HIDDEN_SIZE] [-bhs BACK_HIDDEN_SIZE] [-lr LEARNING_RATE] [-nt {lstm,bi_lstm,conv_1d}] [-vs VALIDATION_SPLIT]
                        [-o {adadelta,adagrad,adam,adamax,ftrl,nadam,rmsprop,sgd}] [-sesm SOUND_EFFECT_START_MARKER] [-seem SOUND_EFFECT_END_MARKER] [-utd] [-d] [-q] [-ver]

Train the Subaligner model

Each subtitle file and its companion audiovisual file need to share the same base filename, the part before the extension.

optional arguments:
  -h, --help            show this help message and exit
  -vd VIDEO_DIRECTORY, --video_directory VIDEO_DIRECTORY
                        Path to the video directory
  -sd SUBTITLE_DIRECTORY, --subtitle_directory SUBTITLE_DIRECTORY
                        Path to the subtitle directory
  -r, --resume          Continue with previous training result if present (hyperparameters passed in will be ignored except for --epochs)
  -dde, --display_done_epochs
                        Display the number of completed epochs
  -utd, --use_training_dump
                        Use training dump instead of files in the video or subtitle directory
  -d, --debug           Print out debugging information
  -q, --quiet           Switch off logging information
  -ver, --version       show program's version number and exit

required arguments:
  -tod TRAINING_OUTPUT_DIRECTORY, --training_output_directory TRAINING_OUTPUT_DIRECTORY
                        Path to the output directory containing training results

optional hyperparameters:
  -bs BATCH_SIZE, --batch_size BATCH_SIZE
                        Number of 32ms samples at each training step
  -do DROPOUT, --dropout DROPOUT
                        Dropout rate between 0 and 1 used at the end of each intermediate hidden layer
  -e EPOCHS, --epochs EPOCHS
                        Total training epochs
  -p PATIENCE, --patience PATIENCE
                        Number of epochs with no improvement after which training will be stopped
  -fhs FRONT_HIDDEN_SIZE, --front_hidden_size FRONT_HIDDEN_SIZE
                        Number of neurons in the front LSTM or Conv1D layer
  -bhs BACK_HIDDEN_SIZE, --back_hidden_size BACK_HIDDEN_SIZE
                        Comma-separated numbers of neurons in the back Dense layers
  -lr LEARNING_RATE, --learning_rate LEARNING_RATE
                        Learning rate of the optimiser
  -nt {lstm,bi_lstm,conv_1d}, --network_type {lstm,bi_lstm,conv_1d}
                        Network type
  -vs VALIDATION_SPLIT, --validation_split VALIDATION_SPLIT
                        Fraction between 0 and 1 of the training data to be used as validation data
  -o {adadelta,adagrad,adam,adamax,ftrl,nadam,rmsprop,sgd}, --optimizer {adadelta,adagrad,adam,adamax,ftrl,nadam,rmsprop,sgd}
                        TensorFlow optimizer
  -sesm SOUND_EFFECT_START_MARKER, --sound_effect_start_marker SOUND_EFFECT_START_MARKER
                        Marker indicating the start of the sound effect which will be ignored during training
  -seem SOUND_EFFECT_END_MARKER, --sound_effect_end_marker SOUND_EFFECT_END_MARKER
                        Marker indicating the end of the sound effect which will be ignored during training
"""

import os
import argparse
import sys
import traceback


def main():
    if sys.version_info.major != 3:
        print("Cannot find Python 3")
        sys.exit(20)
    try:
        import subaligner
        del subaligner
    except ModuleNotFoundError:
        print("Subaligner is not installed")
        sys.exit(20)

    from subaligner._version import __version__
    parser = argparse.ArgumentParser(description="""Train the Subaligner model (%s)\n
Each subtitle file and its companion audiovisual file need to share the same base filename, the part before the extension.""" % __version__,
                                     formatter_class=argparse.RawTextHelpFormatter)
    required_args = parser.add_argument_group("required arguments")
    required_args.add_argument(
        "-tod",
        "--training_output_directory",
        type=str,
        default="",
        help="Path to the output directory containing training results",
        required=True,
    )
    parser.add_argument(
        "-vd",
        "--video_directory",
        type=str,
        default="",
        help="Path to the video directory",
    )
    parser.add_argument(
        "-sd",
        "--subtitle_directory",
        type=str,
        default="",
        help="Path to the subtitle directory",
    )
    parser.add_argument(
        "-r",
        "--resume",
        action="store_true",
        help="Continue with previous training result if present (hyperparameters passed in will be ignored except for --epochs)",
    )
    parser.add_argument(
        "-dde",
        "--display_done_epochs",
        action="store_true",
        help="Display the number of completed epochs",
    )
    parser.add_argument(
        "-sesm",
        "--sound_effect_start_marker",
        type=str,
        default=None,
        help="Marker indicating the start of the sound effect which will be ignored during training",
    )
    parser.add_argument(
        "-seem",
        "--sound_effect_end_marker",
        type=str,
        default=None,
        help="Marker indicating the end of the sound effect which will be ignored during training and used with sound_effect_start_marker",
    )
    hyperparameter_args = parser.add_argument_group("optional hyperparameters")
    hyperparameter_args.add_argument(
        "-bs",
        "--batch_size",
        type=int,
        default=32,
        help="Number of 32ms samples at each training step",
    )
    hyperparameter_args.add_argument(
        "-do",
        "--dropout",
        type=float,
        default=0.2,
        help="Dropout rate between 0 and 1 used at the end of each intermediate hidden layer",
    )
    hyperparameter_args.add_argument(
        "-e",
        "--epochs",
        type=int,
        default=100,
        help="Total training epochs",
    )
    hyperparameter_args.add_argument(
        "-p",
        "--patience",
        type=int,
        default=1000000,
        help="Number of epochs with no improvement after which training will be stopped",
    )
    hyperparameter_args.add_argument(
        "-fhs",
        "--front_hidden_size",
        type=int,
        default=64,
        help="Number of neurons in the front LSTM or Conv1D layer",
    )
    hyperparameter_args.add_argument(
        "-bhs",
        "--back_hidden_size",
        type=str,
        default="32,16",
        help="Comma-separated numbers of neurons in the back Dense layers",
    )
    hyperparameter_args.add_argument(
        "-lr",
        "--learning_rate",
        type=float,
        default=0.001,
        help="Learning rate of the optimiser",
    )
    hyperparameter_args.add_argument(
        "-nt",
        "--network_type",
        type=str,
        choices=["lstm", "bi_lstm", "conv_1d"],
        default="lstm",
        help="Network type",
    )
    hyperparameter_args.add_argument(
        "-vs",
        "--validation_split",
        type=float,
        default=0.25,
        help="Fraction between 0 and 1 of the training data to be used as validation data",
    )
    hyperparameter_args.add_argument(
        "-o",
        "--optimizer",
        type=str,
        choices=["adadelta", "adagrad", "adam", "adamax", "ftrl", "nadam", "rmsprop", "sgd"],
        default="adam",
        help="TensorFlow optimizer",
    )

    parser.add_argument("-utd", "--use_training_dump", action="store_true",
                        help="Use training dump instead of files in the video or subtitle directory")
    parser.add_argument("-d", "--debug", action="store_true",
                        help="Print out debugging information")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="Switch off logging information")
    parser.add_argument("-ver", "--version", action="version", version=__version__)
    FLAGS, unparsed = parser.parse_known_args()

    if FLAGS.training_output_directory == "":
        print("--training_output_directory was not passed in")
        parser.print_usage()
        sys.exit(21)
    if FLAGS.sound_effect_end_marker is not None and FLAGS.sound_effect_start_marker is None:
        print("--sound_effect_start_marker was not passed in when --sound_effect_end_marker was in use")
        parser.print_usage()
        sys.exit(21)

    verbose = FLAGS.debug

    try:
        from subaligner.logger import Logger
        Logger.VERBOSE = FLAGS.debug
        Logger.QUIET = FLAGS.quiet
        from subaligner.exception import UnsupportedFormatException
        from subaligner.exception import TerminalException
        from subaligner.hyperparameters import Hyperparameters
        from subaligner.embedder import FeatureEmbedder
        from subaligner.trainer import Trainer

        output_dir = os.path.abspath(FLAGS.training_output_directory)
        os.makedirs(FLAGS.training_output_directory, exist_ok=True)
        if FLAGS.display_done_epochs:
            output_dir = os.path.abspath(FLAGS.training_output_directory)
            print("Number of epochs done: %s" % str(Trainer.get_done_epochs(os.path.join(output_dir, "training.log"))))
            exit(0)

        model_dir = os.path.abspath(os.path.join(FLAGS.training_output_directory, "models", "training", "model"))
        os.makedirs(model_dir, exist_ok=True)
        weights_dir = os.path.abspath(os.path.join(FLAGS.training_output_directory, "models", "training", "weights"))
        os.makedirs(weights_dir, exist_ok=True)
        config_dir = os.path.abspath(os.path.join(FLAGS.training_output_directory, "models", "training", "config"))
        os.makedirs(config_dir, exist_ok=True)
        video_file_paths = subtitle_file_paths = None
        if not FLAGS.resume:
            if FLAGS.video_directory == "":
                print("--video_directory was not passed in")
                parser.print_usage()
                sys.exit(21)
            if FLAGS.subtitle_directory == "":
                print("--subtitle_directory was not passed in")
                parser.print_usage()
                sys.exit(21)
            video_file_paths = [os.path.abspath(os.path.join(path, p)) for path, _, files in
                                os.walk(FLAGS.video_directory) for p in files if not p.startswith(".")]
            video_file_paths = sorted(video_file_paths, key=lambda x: os.path.splitext(os.path.basename(x))[0])
            subtitle_file_paths = [os.path.abspath(os.path.join(path, p)) for path, _, files in
                                   os.walk(FLAGS.subtitle_directory) for p in files if not p.startswith(".")]
            subtitle_file_paths = sorted(subtitle_file_paths, key=lambda x: os.path.splitext(os.path.basename(x))[0])
        if FLAGS.use_training_dump:
            print("Use data dump from previous training and passed-in video and subtitle directories will be ignored")
            video_file_paths = subtitle_file_paths = None

        hyperparameters = Hyperparameters()
        hyperparameters.batch_size = FLAGS.batch_size
        hyperparameters.dropout = FLAGS.dropout
        hyperparameters.epochs = FLAGS.epochs
        hyperparameters.es_patience = FLAGS.patience
        hyperparameters.front_hidden_size = [FLAGS.front_hidden_size]
        hyperparameters.back_hidden_size = [int(n) for n in FLAGS.back_hidden_size.split(",")]
        hyperparameters.learning_rate = FLAGS.learning_rate
        hyperparameters.network_type = FLAGS.network_type
        hyperparameters.validation_split = FLAGS.validation_split
        hyperparameters.optimizer = FLAGS.optimizer

        trainer = Trainer(FeatureEmbedder())
        trainer.train(video_file_paths,
                      subtitle_file_paths,
                      model_dir,
                      weights_dir,
                      config_dir,
                      output_dir,
                      output_dir,
                      hyperparameters,
                      os.path.join(output_dir, "training.log"),
                      FLAGS.resume,
                      FLAGS.sound_effect_start_marker,
                      FLAGS.sound_effect_end_marker)
    except UnsupportedFormatException as e:
        print(
            "{}\n{}".format(str(e), traceback.format_stack() if verbose else "")
        )
        traceback.print_tb(e.__traceback__)
        sys.exit(23)
    except TerminalException as e:
        print(
            "{}\n{}".format(str(e), traceback.format_stack() if verbose else "")
        )
        traceback.print_tb(e.__traceback__)
        sys.exit(24)
    except Exception as e:
        print(
            "{}\n{}".format(str(e), traceback.format_stack() if verbose else "")
        )
        traceback.print_tb(e.__traceback__)
        sys.exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding=utf8 -*-
#
# stl2srt A program to convert EBU STL subtitle files in the more common SRT format
#
# Copyright 2014 Yann Coupin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
import codecs
import logging
import unicodedata
from xml.etree import ElementTree
import re


class SRT:
    '''A class that behaves like a file object and writes an SRT file'''
    def __init__(self, pathOrFile):
        self.file = open(pathOrFile, 'wb')
        self.counter = 1
        self.file.write(codecs.BOM_UTF8)

    def _formatTime(self, timestamp):
        return "%02u:%02u:%02u,%03u" % (
            timestamp / 3600,
            (timestamp / 60) % 60,
            timestamp % 60,
            (timestamp * 1000) % 1000
        )

    def write(self, start, end, text, encoding):
        text = "\n".join([x for x in text.split("\n") if bool(x)])
        self.file.write(("%0u\n%s --> %s\n%s\n\n" % (self.counter, self._formatTime(start), self._formatTime(end), text)).encode(encoding, errors="replace"))
        self.counter += 1

class iso6937(codecs.Codec):
    '''A class to implement the somewhat exotic iso-6937 encoding which STL files often use'''

    identical = set(range(0x20, 0x7f))
    identical |= set((0xa, 0xa0, 0xa1, 0xa2, 0xa3, 0xa5, 0xa7, 0xab, 0xb0, 0xb1, 0xb2, 0xb3, 0xb5, 0xb6, 0xb7, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf))
    direct_mapping = {
        0x8a: 0x000a, # line break

        0xa8: 0x00a4, # ¤
        0xa9: 0x2018, # ‘
        0xaa: 0x201C, # “
        0xab: 0x00AB, # «
        0xac: 0x2190, # ←
        0xad: 0x2191, # ↑
        0xae: 0x2192, # →
        0xaf: 0x2193, # ↓

        0xb4: 0x00D7, # ×
        0xb8: 0x00F7, # ÷
        0xb9: 0x2019, # ’
        0xba: 0x201D, # ”
        0xbc: 0x00BC, # ¼
        0xbd: 0x00BD, # ½
        0xbe: 0x00BE, # ¾
        0xbf: 0x00BF, # ¿

        0xd0: 0x2015, # ―
        0xd1: 0x00B9, # ¹
        0xd2: 0x00AE, # ®
        0xd3: 0x00A9, # ©
        0xd4: 0x2122, # ™
        0xd5: 0x266A, # ♪
        0xd6: 0x00AC, # ¬
        0xd7: 0x00A6, # ¦
        0xdc: 0x215B, # ⅛
        0xdd: 0x215C, # ⅜
        0xde: 0x215D, # ⅝
        0xdf: 0x215E, # ⅞

        0xe0: 0x2126, # Ohm Ω
        0xe1: 0x00C6, # Æ
        0xe2: 0x0110, # Đ
        0xe3: 0x00AA, # ª
        0xe4: 0x0126, # Ħ
        0xe6: 0x0132, # Ĳ
        0xe7: 0x013F, # Ŀ
        0xe8: 0x0141, # Ł
        0xe9: 0x00D8, # Ø
        0xea: 0x0152, # Œ
        0xeb: 0x00BA, # º
        0xec: 0x00DE, # Þ
        0xed: 0x0166, # Ŧ
        0xee: 0x014A, # Ŋ
        0xef: 0x0149, # ŉ

        0xf0: 0x0138, # ĸ
        0xf1: 0x00E6, # æ
        0xf2: 0x0111, # đ
        0xf3: 0x00F0, # ð
        0xf4: 0x0127, # ħ
        0xf5: 0x0131, # ı
        0xf6: 0x0133, # ĳ
        0xf7: 0x0140, # ŀ
        0xf8: 0x0142, # ł
        0xf9: 0x00F8, # ø
        0xfa: 0x0153, # œ
        0xfb: 0x00DF, # ß
        0xfc: 0x00FE, # þ
        0xfd: 0x0167, # ŧ
        0xfe: 0x014B, # ŋ
        0xff: 0x00AD, # Soft hyphen
    }
    diacritic = {
        0xc1: 0x0300, # grave accent
        0xc2: 0x0301, # acute accent
        0xc3: 0x0302, # circumflex
        0xc4: 0x0303, # tilde
        0xc5: 0x0304, # macron
        0xc6: 0x0306, # breve
        0xc7: 0x0307, # dot
        0xc8: 0x0308, # umlaut
        0xca: 0x030A, # ring
        0xcb: 0x0327, # cedilla
        0xcd: 0x030B, # double acute accent
        0xce: 0x0328, # ogonek
        0xcf: 0x030C, # caron
    }


    def decode(self, input):
        output = []
        state = None
        count = 0
        for char in input:
            char = ord(char)
            # End of a subtitle text
            count += 1
            if not state and char in self.identical:
                output.append(char)
            elif not state and char in self.direct_mapping:
                output.append(self.direct_mapping[char])
            elif not state and char in self.diacritic:
                state = self.diacritic[char]
            elif state:
                combined = unicodedata.normalize('NFC', chr(char) + chr(state))
                if combined and len(combined) == 1:
                    output.append(ord(combined))
                state = None
        return (''.join(map(chr, output)), len(input))

    def search(self, name):
        if name in ('iso6937', 'iso_6937-2'):
            return codecs.CodecInfo(self.encode, self.decode, name='iso_6937-2')

    def encode(self, input):
        pass

codecs.register(iso6937().search)

class RichText:

    def __init__(self, use_html_tags):
        self.tag_stack = []
        self.opened_tags = set()
        self.output = []
        self.add_html_tags = use_html_tags

    def write(self, string):
        self.output.append(string)

    def openTag(self, tag_name, tag_html=None):
        if not tag_html:
            tag_html = '<%s>' % tag_name
        if tag_name not in self.opened_tags:
            self.tag_stack.append((tag_name, tag_html))
            self.opened_tags.add(tag_name)
            if not self.add_html_tags:
                self.output.append(' ')
            else:
                self.output.append(tag_html)

    def closeTag(self, tag):
        if not self.add_html_tags:
            return
        tag_html = '</%s>' % tag
        if tag in self.opened_tags:
            reopen_stack = []
            while self.tag_stack:
                tag_to_close = self.tag_stack.pop()
                if tag_to_close[0] == tag:
                    self.output.append(tag_html)
                    self.opened_tags.remove(tag)
                    break
                else:
                    reopen_stack += tag_to_close
            for tag_to_reopen in reopen_stack:
                self.output.append(tag_to_reopen[1])
                self.tag_stack.append(tag_to_reopen)

    def __str__(self):
        if not self.add_html_tags:
            return ''.join(self.output)

        closing_tags = []
        # Close all the tags still open
        for tag in self.tag_stack[::-1]:
            closing_tags.append('</%s>' % tag[0])
        return ''.join(self.output + closing_tags)

class STL:
    '''A class that behaves like a file object and reads an STL file'''

    GSIfields = 'CPN DFC DSC CCT LC OPT OET TPT TET TN TCD SLR CD RD RN TNB TNS TNG MNC MNR TCS TCP TCF TND DSN CO PUB EN ECD UDA'.split(' ')
    TTIfields = 'SGN SN EBN CS TCIh TCIm TCIs TCIf TCOh TCOm TCOs TCOf VP JC CF TF'.split(' ')

    def __init__(self, pathOrFile, richFormatting=False):
        self.file = open(pathOrFile, 'rb')

        self.richFormatting = richFormatting

        self._readGSI()

    def __bcdTimestampDecode(self, timestamp):
        # Special case for people that can't bother to read a spec
        if timestamp == '________':
            return 0.0

        # BCD coded time with limited significant bits as per EBU Tech. 3097-E
        safe_bytes = [x[0]&x[1] for x in zip((0x2, 0xf, 0x7, 0xf, 0x7, 0xf, 0x3, 0xf), struct.unpack('8B', timestamp))]
        return sum([x[0]*x[1] for x in zip((36000, 3600, 600, 60, 10, 1, 10.0 / self.fps, 1.0 / self.fps), safe_bytes)])

    def _readGSI(self):
        self.GSI = dict(list(zip(
            self.GSIfields,
            struct.unpack('3s8sc2s2s32s32s32s32s32s32s16s6s6s2s5s5s3s2s2s1s8s8s1s1s3s32s32s32s75x576s', self.file.read(1024))
        )))
        GSI = self.GSI
        # GSI ={k: v.decode("utf-8")  for k, v in self.GSI.items()}
        logging.debug(GSI)
        #self.gsiCodePage = 'cp%s' % GSI['CPN']
        if GSI['DFC'] == b'STL24.01':
            self.fps = 24
        elif GSI['DFC'] == b'STL25.01':
            self.fps = 25
        elif GSI['DFC'] == b'STL30.01':
            self.fps = 30
        else:
            raise Exception('Invalid DFC')
        self.codePage = {
            '00': 'iso_6937-2',
            '01': 'iso-8859-5',
            '02': 'iso-8859-6',
            '03': 'iso-8859-7',
            '04': 'iso-8859-8',
        }[GSI['CCT'].decode('utf-8')]
        self.numberOfTTI = int(GSI['TNB'])
        if GSI['TCS'] == '1':
            # BCD coded time with limited significant bits

            self.startTime = self.__bcdTimestampDecode(GSI['TCP'])
        else:
            self.startTime = 0.0
        logging.debug(self.__dict__)

    def __timecodeDecode(self, h, m, s, f):
        return 3600 * h + 60 * m + s + float(f) / self.fps

    def __parseFormatting(self, text, addHtmlTags):
        colorCodes = [
            '#000000', # black
            '#ff0000', # red
            '#00ff00', # green
            '#ffff00', # yellow
            '#0000ff', # blue
            '#ff00ff', # magenta
            '#00ffff', # cyan
            '#ffffff', # white
        ]
        currentColor = 7 # White is the default color
        output = RichText(addHtmlTags)

        first_line = True

        for char in "".join(map(chr, text)):
            ochar = ord(char)
            if ochar == 0x80:
                output.openTag('i')
            elif ochar == 0x81:
                output.closeTag('i')
            elif ochar == 0x82:
                output.openTag('u')
            elif ochar == 0x83:
                output.closeTag('u')
            elif ochar == 0xe:
                output.openTag('b')
            elif ochar == 0xc:
                output.closeTag('b')
            elif ochar in (0,1,2,3,4,5,6,7,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17):
                color = ochar % 0x10
                if color != currentColor:
                    currentColor = color
                    output.closeTag('font')
                    output.openTag('font', '<font color="%s">' % colorCodes[color])
            elif ochar == 0x8a and first_line:
                output.write("\n")
                first_line = False
            elif ochar == 0x8f:
                break
            elif (ochar & 0x7F) >= 0x20:
                output.write(char)

        return str(output)

    def _readTTI(self):
        while (True):
            tci = None
            tco = None
            txt = []

            while (True):
                data = self.file.read(128)
                if not data:
                    raise StopIteration()
                TTI = dict(list(zip(
                    self.TTIfields,
                    struct.unpack('<BHBBBBBBBBBBBBB112s', data)
                )))
                logging.debug(TTI)
                # if comment skip
                if TTI['CF']:
                    continue
                # discard blocks with user data
                if TTI['EBN'] == 254:
                    continue
                if not tci:
                    tci = self.__timecodeDecode(TTI['TCIh'], TTI['TCIm'], TTI['TCIs'], TTI['TCIf']) - self.startTime
                    tco = self.__timecodeDecode(TTI['TCOh'], TTI['TCOm'], TTI['TCOs'], TTI['TCOf']) - self.startTime
                text = TTI['TF']
                text = self.__parseFormatting(text, self.richFormatting)
                text = text.strip()
                txt += text
                if TTI['EBN'] == 255:
                    # skip empty subtitles and those before the start of the show
                    if txt and tci >= 0:
                        return (tci, tco, ''.join(txt))
                    break

    def __iter__(self):
        return self

    def __next__(self):
        return self._readTTI()

class TT:
    '''A class that behaves like a file object and reads a TT subtitle file'''

    __named_colors = {
        'black': '#000000',
        'silver': '#c0c0c0',
        'gray': '#808080',
        'white': '#ffffff',
        'maroon': '#800000',
        'red': '#ff0000',
        'purple': '#800080',
        'fuchsia': '#ff00ff',
        'magenta': '#ff00ff',
        'green': '#008000',
        'lime': '#00ff00',
        'olive': '#808000',
        'yellow': '#ffff00',
        'navy': '#000080',
        'blue': '#0000ff',
        'teal': '#008080',
        'aqua': '#00ffff',
        'cyan': '#00ffff',
    }

    def __init__(self, source, richFormatting):
        self.source = source

        self.richFormatting = richFormatting
        self.frameRate = 30.0
        self.frameRateMultiplier = 1.0
        self.subFrameRate = 1.0
        self.tickRate = None

        self._parseXML()

    def __parse_style(self, element):
        style = {}
        for k, v in list(element.items()):
            if k == '{tts}color':
                if re.match(r'#\d{6}\d{2}?', v):
                    style['color'] = v[0:7]
                else:
                    rgb_color = re.match(r'rgb\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\)', v.strip())
                    if rgb_color:
                        style['color'] = '#%02x%02x%02x' % (int(rgb_color.group(1)), int(rgb_color.group(2)), int(rgb_color.group(3)))
                    else:
                        c = self.__named_colors.get(v)
                        if c:
                            style['color'] = c
            elif k == '{tts}fontStyle':
                style['italic'] = v == 'italic'
            elif k == '{tts}fontWeight':
                style['bold'] = v == 'bold'
            elif k == '{tts}textDecoration':
                style['underline'] = v == 'underline'
        return style

    def __process_time(self, text):
        coefs = [3600, 60, 1]
        time = 0.0

        offset_match = re.match(r'(\d+)(:?\.\d+)?(h|m|s|ms|f|t)', text)
        if offset_match:
            return float(offset_match.group(1)) * {
                'h': 3600.0,
                'm': 60.0,
                's': 1.0,
                'ms': 0.001,
                'f': 1.0/(self.frameRate * self.frameRateMultiplier),
                't': 1.0/self.tickRate
            }.get(offset_match.group(2), 1.0)
        params = text.split(':')
        if len(params) in (3, 4):
            if len(params) == 4:
                frames = params[3].split('.', 2)
                if len(frames) == 1:
                    params[2] = float(params[2]) + float(params[3]) / (self.frameRate * self.frameRateMultiplier)
                else:
                    params[2] = float(params[2]) + (
                        float(frames[0]) / self.frameRate +
                        float(frames[1]) / (self.frameRate * self.subFrameRate)
                    ) * self.frameRateMultiplier
                del params[3]
            for c, v in zip(coefs, params):
                time += c*float(v)
            return time
        return 0.0

    def _parseXML(self):
        # Normalize namespaces to a single alias. The draft namespace are still used in some file which makes searching for tags cumbersome
        namespace_clean = {
            'http://www.w3.org/2006/10/ttaf1': 'tt',
            'http://www.w3.org/2006/04/ttaf1': 'tt',
            'http://www.w3.org/ns/ttml': 'tt',
            'http://www.w3.org/2006/10/ttaf1#styling': 'tts',
            'http://www.w3.org/2006/04/ttaf1#styling': 'tts',
            'http://www.w3.org/ns/ttml#styling': 'tts',
            'http://www.w3.org/2006/10/ttaf1#parameter': 'ttp',
            'http://www.w3.org/2006/04/ttaf1#parameter': 'ttp',
            'http://www.w3.org/ns/ttml#parameter': 'ttp',
        }
        def normalize_qname(name):
            if name[0] == '{':
                (ns, name) = name[1:].split('}', 1)
                ns = namespace_clean.get(ns, ns)
                return '{%s}%s' % (ns, name)
            return name

        xml = ElementTree.parse(self.source)
        for element in xml.getiterator():
            element.tag = normalize_qname(element.tag)
            for k, v in list(element.items()):
                new_k = normalize_qname(k)
                if k != new_k:
                    del element.attrib[k]
                    element.attrib[new_k] = v

        # Define style aliases
        styles = {}
        regions = {}

        root = xml.getiterator()[0]
        if int(root.get('{ttp}tickRate', 0)) > 0:
            self.tickRate = int(root.get('{ttp}tickRate'))
        if int(root.get('{ttp}frameRate', 0)) > 0:
            self.frameRate = int(root.get('{ttp}frameRate'))
        if int(root.get('{ttp}subFrameRate', 0)) > 0:
            self.subFrameRate = int(root.get('{ttp}subFrameRate'))
        if root.get('{ttp}frameRateMultiplier'):
            num, denom = root.get('{ttp}frameRateMultiplier').split(' ')
            self.frameRateMultiplier = float(num) / float(denom)
        if not self.tickRate:
            self.tickRate = self.frameRate * self.subFrameRate * self.frameRateMultiplier

        # Build a cache for the default styles
        for style_tag in xml.findall('{tt}head/{tt}styling/{tt}style'):
            style = self.__parse_style(style_tag)
            styles[style_tag.get('{http://www.w3.org/XML/1998/namespace}id')] = style

        # Build a cache for the default style of the regions
        for region_tag in xml.findall('{tt}head/{tt}layout/{tt}region'):
            region = self.__parse_style(region_tag)
            regions[region_tag.get('{http://www.w3.org/XML/1998/namespace}id')] = region

        def compute_style_tree(element):
            style_ref = element.get('style')
            region_ref = element.get('region')

            style = {}
            if region_ref:
                style.update(regions[region_ref])
            if style_ref:
                style.update(styles[style_ref])
            style.update(self.__parse_style(element))

            return style

        def styleToHtml(tag, value):
            return {
                'bold': ('b', '<b>'),
                'italic': ('i', '<i>'),
                'underline': ('u', '<u>'),
                'color': ('font', '<font color="%s">' % value),
            }[tag]

        def openTags(output, style_pairs):
            (before, after) = style_pairs
            for tag in sorted(after.keys()):
                new_value = after[tag]
                old_value = before.get(tag, None)
                if old_value == None and new_value:
                    html = styleToHtml(tag, new_value)
                    output.openTag(html[0], html[1])
                elif old_value != new_value:
                    if new_value:
                        html = styleToHtml(tag, new_value)
                        output.openTag(html[0], html[1])
                    else:
                        output.closeTag(styleToHtml(tag, new_value)[0])

        def closeTags(output, style_pairs):
            (before, after) = style_pairs
            for tag in sorted(list(after.keys()), reverse=True):
                new_value = after[tag]
                old_value = before.get(tag, None)
                if old_value == None and new_value:
                    output.closeTag(styleToHtml(tag, new_value)[0])
                elif old_value != new_value:
                    if new_value:
                        output.closeTag(styleToHtml(tag, new_value)[0])
                    else:
                        html = styleToHtml(tag, before[tag])
                        output.openTag(html[0], html[1])

        # Store the subs in a list
        self.subs = []
        prev_sub = None
        content = None
        sub_grouping = False
        for sub in xml.findall('{tt}body/{tt}div/{tt}p'):
            begin = self.__process_time(sub.get('begin'))
            if not sub.get('end'):
                end = begin + self.__process_time(sub.get('dur'))
            else:
                end = self.__process_time(sub.get('end'))

            style_stack = [{'color': '#ffffff'}] # default color

            if not prev_sub or begin != prev_sub[0] or end != prev_sub[1]:
                content = RichText(self.richFormatting)
                sub_grouping = False
            else:
                content.write("\n")
                sub_grouping = True

            def parseChildTree(element_list):
                for child in element_list:
                    style_stack.append(compute_style_tree(child))
                    openTags(content, style_stack[-2:])
                    if child.text and child.text.strip():
                        content.write(child.text.strip())
                    if child.tag == '{tt}br':
                        content.write("\n")
                    parseChildTree(child.getchildren())
                    if child.tail and child.tail.strip():
                        content.write(child.tail.strip())
                    closeTags(content, style_stack[-2:])
                    style_stack.pop()

            parseChildTree([sub])

            # try to regroup subtitles if possible
            if sub_grouping:
                self.subs[-1][2] = str(content)
            else:
                prev_sub = [begin, end, str(content)]
                self.subs.append(prev_sub)

    def __iter__(self):
        return iter(self.subs)

if __name__ == '__main__':

    from optparse import OptionParser
    import sys

    parser = OptionParser(usage = 'usage: %prog [options] input output')
    parser.set_defaults(reader_class=STL)
    parser.add_option('-d', '--debug', dest='debug_level', action='store_const', const=logging.DEBUG, default=logging.ERROR)
    parser.add_option('-r', '--rich', dest='rich_formatting', action='store_true', default=False, help='Output text with some formatting, the following HTML tags are used: b i u font(color)')
    parser.add_option("-s", "--stl", dest="reader_class", action="store_const", const=STL,
                            help="Set input file format as STL (default)")
    parser.add_option("-t", "--tt", dest="reader_class", action="store_const", const=TT,
                            help="Set input file format as TT, handles both the EBU and SMPTE variants")

    (options, args) = parser.parse_args()

    if len(args) != 2:
        parser.print_help()
        sys.exit(1)

    logging.basicConfig(level=options.debug_level)

    input = options.reader_class(args[0], options.rich_formatting)
    c = SRT(args[1])
    for sub in input:
        (tci, tco, txt) = sub
        c.write(tci, tco, txt)
    c.file.close()



class UnsupportedFormatException(Exception):
    """ An exception raised due to unsupported formats."""


class TerminalException(Exception):
    """ An exception raised due to unrecoverable failures."""


class NoFrameRateException(Exception):
    """ An exception raised due to frame rate not found."""



import sys
import json
from typing import Any, List


class Hyperparameters(object):
    """ The configuration on hyperparameters used for training
    """

    OPTIMIZERS = ["adadelta", "adagrad", "adam", "adamax", "ftrl", "nadam", "rmsprop", "sgd"]

    def __init__(self) -> None:
        """Hyperparameters initialiser setting default values"""

        self.__learning_rate = 0.001
        self.__hidden_size = {
            "front_layers": [64],
            "back_layers": [32, 16]
        }
        self.__dropout = 0.2
        self.__epochs = 100
        self.__optimizer = "Adam"
        self.__loss = "binary_crossentropy"
        self.__metrics = ["accuracy"]
        self.__batch_size = 32
        self.__validation_split = 0.25
        self.__monitor = "val_loss"
        self.__es_mode = "min"
        self.__es_min_delta = 0.00001
        self.__es_patience = sys.maxsize
        self.__network_type = "lstm"

    def __eq__(self, other: Any) -> bool:
        """Comparator for Hyperparameters objects"""

        if isinstance(other, Hyperparameters):
            return all([
                self.__learning_rate == other.learning_rate,
                self.__hidden_size["front_layers"] == other.front_hidden_size,
                self.__hidden_size["back_layers"] == other.back_hidden_size,
                self.__dropout == other.dropout,
                self.__epochs == other.epochs,
                self.__optimizer == other.optimizer,
                self.__loss == other.loss,
                self.__metrics == other.metrics,
                self.__batch_size == other.batch_size,
                self.__validation_split == other.validation_split,
                self.__monitor == other.monitor,
                self.__es_mode == other.es_mode,
                self.__es_min_delta == other.es_min_delta,
                self.__es_patience == other.es_patience,
                self.__network_type == other.network_type
            ])
        return False

    @property
    def learning_rate(self) -> float:
        return self.__learning_rate

    @learning_rate.setter
    def learning_rate(self, value: float) -> None:
        self.__learning_rate = value

    @property
    def front_hidden_size(self) -> List[int]:
        return self.__hidden_size["front_layers"]

    @front_hidden_size.setter
    def front_hidden_size(self, value: List[int]) -> None:
        self.__hidden_size["front_layers"] = value

    @property
    def back_hidden_size(self) -> List[int]:
        return self.__hidden_size["back_layers"]

    @back_hidden_size.setter
    def back_hidden_size(self, value: List[int]) -> None:
        self.__hidden_size["back_layers"] = value

    @property
    def dropout(self) -> float:
        return self.__dropout

    @dropout.setter
    def dropout(self, value: float) -> None:
        self.__dropout = value

    @property
    def epochs(self) -> int:
        return self.__epochs

    @epochs.setter
    def epochs(self, value: int) -> None:
        self.__epochs = value

    @property
    def optimizer(self) -> str:
        return self.__optimizer

    @optimizer.setter
    def optimizer(self, value: str) -> None:
        if value not in self.OPTIMIZERS:
            raise ValueError("Optimizer {} is not supported".format(value))

        if value.lower() == "adadelta":
            self.__optimizer = "Adadelta"
        elif value.lower() == "adagrad":
            self.__optimizer = "Adagrad"
        elif value.lower() == "adam":
            self.__optimizer = "Adam"
        elif value.lower() == "adamax":
            self.__optimizer = "Adamax"
        elif value.lower() == "ftrl":
            self.__optimizer = "Ftrl"
        elif value.lower() == "nadam":
            self.__optimizer = "Nadam"
        elif value.lower() == "rmsprop":
            self.__optimizer = "RMSprop"
        elif value.lower() == "sgd":
            self.__optimizer = "SGD"

    @property
    def loss(self) -> str:
        return self.__loss

    @property
    def metrics(self) -> List[str]:
        return self.__metrics

    @metrics.setter
    def metrics(self, value: List[str]):
        self.__metrics = value

    @property
    def batch_size(self) -> int:
        return self.__batch_size

    @batch_size.setter
    def batch_size(self, value: int) -> None:
        self.__batch_size = value

    @property
    def validation_split(self) -> float:
        return self.__validation_split

    @validation_split.setter
    def validation_split(self, value: float) -> None:
        self.__validation_split = value

    @property
    def monitor(self) -> str:
        return self.__monitor

    @monitor.setter
    def monitor(self, value: str) -> None:
        self.__monitor = value

    @property
    def es_mode(self) -> str:
        return self.__es_mode

    @es_mode.setter
    def es_mode(self, value: str) -> None:
        self.__es_mode = value

    @property
    def es_min_delta(self) -> float:
        return self.__es_min_delta

    @es_min_delta.setter
    def es_min_delta(self, value: float) -> None:
        self.__es_min_delta = value

    @property
    def es_patience(self) -> int:
        return self.__es_patience

    @es_patience.setter
    def es_patience(self, value: int) -> None:
        self.__es_patience = value

    @property
    def network_type(self) -> str:
        return self.__network_type

    @network_type.setter
    def network_type(self, value: str) -> None:
        self.__network_type = value

    def to_json(self) -> str:
        """Serialise hyperparameters into JSON string

        Returns:
            string -- The serialised hyperparameters in JSON
        """
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def to_file(self, file_path: str) -> None:
        """Serialise hyperparameters into JSON and save the content to a file

        Arguments:
            file_path {string} -- The path to the file containing saved hyperparameters.
        """
        with open(file_path, "w", encoding="utf8") as file:
            file.write(self.to_json())

    def clone(self) -> "Hyperparameters":
        """Make a cloned hyperparameters object

        Returns:
            Hyperparameters -- The cloned Hyperparameters object.
        """
        return self.from_json(self.to_json())

    @classmethod
    def from_json(cls, json_str: str) -> "Hyperparameters":
        """Deserialise JSON string into a Hyperparameters object

        Arguments:
            json_str {string} -- Hyperparameters in JSON.

        Returns:
            Hyperparameters -- The deserialised Hyperparameters object.
        """
        hp = cls()
        hp.__dict__ = json.loads(json_str)
        return hp

    @classmethod
    def from_file(cls, file_path: str) -> "Hyperparameters":
        """Deserialise a file content into a Hyperparameters object

        Arguments:
            file_path {string} -- The path to the file containing hyperparameters.

        Returns:
            Hyperparameters -- The deserialised Hyperparameters object.
        """
        with open(file_path, "r", encoding="utf8") as file:
            return cls.from_json(file.read())



import logging
from typing import Dict
from absl import logging as absl_logging
from .singleton import Singleton
absl_logging._warn_preinit_stderr = 0


class Logger(Singleton):
    """Common logging."""

    VERBOSE = True
    QUIET = False

    def __init__(self, output_log: str = "output.log") -> None:
        self.__loggers: Dict[str, logging.Logger] = {}
        self.__output_log = output_log

    def get_logger(self, name: str) -> logging.Logger:

        if self.__loggers.get(name) is not None:
            return self.__loggers.get(name)  # type: ignore
        else:
            logger = logging.getLogger(name)
            logger.setLevel(logging.INFO)
            if Logger.VERBOSE:
                logger.setLevel(logging.DEBUG)
            if Logger.QUIET:
                logger.setLevel(logging.ERROR)
            formatter = logging.Formatter(
                "%(name)s - %(levelname)s - %(threadName)-9s - %(message)s"
            )

            file_handler = logging.FileHandler(self.__output_log, "w+")
            logger.propagate = False
            file_handler.setFormatter(formatter)

            console_handler = logging.StreamHandler()
            console_handler.setFormatter(formatter)

            if not logger.handlers:
                logger.addHandler(file_handler)
                logger.addHandler(console_handler)

            self.__loggers[name] = logger
            return logger



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import setup

with open(os.path.join(os.getcwd(), "subaligner", "_version.py")) as f:
    exec(f.read())

with open("README.md") as readme_file:
    readme = readme_file.read()

if "STRETCH_OFF" not in os.environ:
    with open("requirements.txt") as requirements_file:
        requirements = requirements_file.read().splitlines()[::-1]
else:
    with open("requirements-app.txt") as requirements_file:
        requirements = requirements_file.read().splitlines()[::-1]

setup(name="subaligner",
      version=__version__,
      author="Xi Bai",
      author_email="xi.bai.ed@gmail.com",
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9"
      ],
      license="MIT",
      url="https://subaligner.readthedocs.io/en/latest/",
      description="Automatically synchronize subtitles to audiovisual content and translate them into various languages with pretrained deep neural networks and forced alignments.",
      long_description=readme + "\n\n",
      long_description_content_type='text/markdown',
      python_requires=">=3.6",
      package_dir={"subaligner": "subaligner"},
      packages=[
          "subaligner",
          "subaligner.lib",
          "subaligner.subaligner_1pass",
          "subaligner.subaligner_2pass",
          "subaligner.subaligner_batch",
          "subaligner.subaligner_convert",
          "subaligner.subaligner_train",
          "subaligner.subaligner_tune",
          "subaligner.models.training.model",
          "subaligner.models.training.weights",
          "subaligner.models.training.config",
      ],
      package_data={
          "subaligner.models.training.model": ["model.hdf5"],
          "subaligner.models.training.weights": ["weights.hdf5"],
          "subaligner.models.training.config": ["hyperparameters.json"],
      },
      install_requires=requirements,
      test_suite="tests.subaligner",
      setup_requires=["numpy>=1.14.1,<1.18.0"],
      scripts=[
          "bin/subaligner",
          "bin/subaligner_1pass",
          "bin/subaligner_2pass",
          "bin/subaligner_batch",
          "bin/subaligner_convert",
          "bin/subaligner_train",
          "bin/subaligner_tune",
      ],
      entry_points={
          "console_scripts": [
              "subaligner=subaligner.__main__:main",
              "subaligner_1pass=subaligner.subaligner_1pass.__main__:main",
              "subaligner_2pass=subaligner.subaligner_2pass.__main__:main",
              "subaligner_batch=subaligner.subaligner_batch.__main__:main",
              "subaligner_convert=subaligner.subaligner_convert.__main__:main",
              "subaligner_train=subaligner.subaligner_train.__main__:main",
              "subaligner_tune=subaligner.subaligner_tune.__main__:main",
          ]})



import os
import subprocess
import pysubs2
import requests
import shutil
import cchardet
import shlex

from pycaption import (
    CaptionConverter,
    SRTWriter,
    SRTReader,
    DFXPWriter,
    DFXPReader,
    SAMIWriter,
    SAMIReader,
    SCCWriter,
    SCCReader,
)

from captionstransformer.core import Caption
from captionstransformer.core import Reader
from captionstransformer.sbv import Reader as SbvReader, Writer as SbvWriter
from captionstransformer.srt import Reader as SrtReader, Writer as SrtWriter
from captionstransformer.transcript import Reader as TranscriptReader, Writer as TranscriptWriter
from bs4 import BeautifulSoup
from aeneas.language import Language
from datetime import datetime
from typing import Optional, TextIO, BinaryIO, Union, Callable, Any, Tuple, List, Dict
from .exception import TerminalException
from subaligner.lib.to_srt import STL, SRT


class Utils(object):
    """Utility functions
    """

    FFMPEG_BIN = os.getenv("FFMPEG_PATH") or os.getenv("ffmpeg_path") or "ffmpeg"

    @staticmethod
    def srt2ttml(srt_file_path: str, ttml_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to TTML subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            ttml_file_path {string} -- The path to the TTML file.
        """

        file: Union[TextIO, BinaryIO]
        converter = CaptionConverter()
        encoding = Utils.detect_encoding(srt_file_path)
        with open(srt_file_path, "r", encoding=encoding) as file:
            converter.read(file.read(), SRTReader())
        if ttml_file_path is None:
            ttml_file_path = srt_file_path.replace(".srt", ".xml")
        with open(ttml_file_path, "wb") as file:
            file.write(converter.write(DFXPWriter()).encode(encoding, errors="replace"))

    @staticmethod
    def ttml2srt(ttml_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert TTML subtitles to SubRip subtitles.

        Arguments:
            ttml_file_path {string} -- The path to the TTML file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        file: Union[TextIO, BinaryIO]
        converter = CaptionConverter()
        encoding = Utils.detect_encoding(ttml_file_path)
        with open(ttml_file_path, "r", encoding=encoding) as file:
            converter.read(file.read(), DFXPReader())
        if srt_file_path is None:
            srt_file_path = ttml_file_path.replace(".xml", ".srt")
        with open(srt_file_path, "wb") as file:
            file.write(converter.write(SRTWriter()).encode(encoding, errors="replace"))

    @staticmethod
    def srt2vtt(srt_file_path: str, vtt_file_path: Optional[str] = None, timeout_secs: int = 30) -> None:
        """Convert SubRip subtitles to WebVTT subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            vtt_file_path {string} -- The path to the WebVTT file.
            timeout_secs {int} -- The timeout in seconds on conversion {default: 30}.
        """

        _vtt_file_path = srt_file_path.replace(".srt", ".vtt") if vtt_file_path is None else vtt_file_path
        encoding = Utils.detect_encoding(srt_file_path)
        command = "{0} -y -sub_charenc {1} -i '{2}' -f webvtt '{3}'".format(Utils.FFMPEG_BIN, encoding, srt_file_path, _vtt_file_path)
        timeout_msg = "Timeout on converting SubRip to WebVTT: {}".format(srt_file_path)
        error_msg = "Cannot convert SubRip to WebVTT: {}".format(srt_file_path)

        def _callback(returncode: int, std_err: str) -> None:
            if returncode != 0:
                raise TerminalException(
                    "Cannot convert SubRip to WebVTT: {} with error {}".format(
                        srt_file_path, std_err
                    )
                )
            Utils.remove_trailing_newlines(_vtt_file_path, encoding)

        Utils._run_command(command, timeout_secs, timeout_msg, error_msg, _callback)

    @staticmethod
    def vtt2srt(vtt_file_path: str, srt_file_path: Optional[str] = None, timeout_secs: int = 30) -> None:
        """Convert WebVTT subtitles to SubRip subtitles.

        Arguments:
            vtt_file_path {string} -- The path to the WebVTT file.
            srt_file_path {string} -- The path to the SubRip file.
            timeout_secs {int} -- The timeout in seconds on conversion {default: 30}.
        """

        _srt_file_path = vtt_file_path.replace(".vtt", ".srt") if srt_file_path is None else srt_file_path
        encoding = Utils.detect_encoding(vtt_file_path)
        command = "{0} -y -sub_charenc {1} -i '{2}' -f srt '{3}'".format(Utils.FFMPEG_BIN, encoding, vtt_file_path, _srt_file_path)
        timeout_msg = "Timeout on converting WebVTT to SubRip: {}".format(vtt_file_path)
        error_msg = "Cannot convert WebVTT to SubRip: {}".format(vtt_file_path)

        def _callback(returncode: int, std_err: str) -> None:
            if returncode != 0:
                raise TerminalException(
                    "Cannot convert WebVTT to SubRip: {} with error {}".format(
                        vtt_file_path, std_err
                    )
                )
            Utils.remove_trailing_newlines(_srt_file_path, encoding)

        Utils._run_command(command, timeout_secs, timeout_msg, error_msg, _callback)

    @staticmethod
    def srt2ass(srt_file_path: str, ass_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to Advanced SubStation Alpha v4.0+ subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            ass_file_path {string} -- The path to the ASS file.
        """

        new_ass_file_path, encoding = Utils.__convert_subtitle(srt_file_path, "srt", ass_file_path, "ass", "ass")
        Utils.remove_trailing_newlines(new_ass_file_path, encoding)

    @staticmethod
    def ass2srt(ass_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert Advanced SubStation Alpha v4.0+ subtitles to SubRip subtitles.

        Arguments:
            ass_file_path {string} -- The path to the ASS file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        new_srt_file_path, encoding = Utils.__convert_subtitle(ass_file_path, "ass", srt_file_path, "srt", "srt")
        Utils.remove_trailing_newlines(new_srt_file_path, encoding)

    @staticmethod
    def srt2ssa(srt_file_path: str, ssa_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to SubStation Alpha v4.0 subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            ssa_file_path {string} -- The path to the SSA file.
        """

        new_ssa_file_path, encoding = Utils.__convert_subtitle(srt_file_path, "srt", ssa_file_path, "ssa", "ssa")
        Utils.remove_trailing_newlines(new_ssa_file_path, encoding)

    @staticmethod
    def ssa2srt(ssa_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert SubStation Alpha v4.0 subtitles to SubRip subtitles.

        Arguments:
            ssa_file_path {string} -- The path to the SSA file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        new_srt_file_path, encoding = Utils.__convert_subtitle(ssa_file_path, "ssa", srt_file_path, "srt", "srt")
        Utils.remove_trailing_newlines(new_srt_file_path, encoding)

    @staticmethod
    def srt2microdvd(srt_file_path: str, microdvd_file_path: Optional[str] = None, frame_rate: Optional[float] = 25.0):
        """Convert SubRip subtitles to MicroDVD subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            microdvd_file_path {string} -- The path to the MicroDVD file.
            frame_rate {float} -- The frame rate for frame-based MicroDVD.
        """

        new_microdvd_file_path, encoding = Utils.__convert_subtitle(srt_file_path, "srt", microdvd_file_path, "sub", "microdvd", frame_rate=frame_rate)
        Utils.remove_trailing_newlines(new_microdvd_file_path, encoding)

    @staticmethod
    def microdvd2srt(microdvd_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert MicroDVD subtitles to SubRip subtitles.

        Arguments:
            microdvd_file_path {string} -- The path to the MPL2 file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        new_srt_file_path, encoding = Utils.__convert_subtitle(microdvd_file_path, "sub", srt_file_path, "srt", "srt")
        Utils.remove_trailing_newlines(new_srt_file_path, encoding)

    @staticmethod
    def srt2mpl2(srt_file_path: str, mpl2_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to MPL2 subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            mpl2_file_path {string} -- The path to the MPL2 file.
        """

        new_mpl2_file_path, encoding = Utils.__convert_subtitle(srt_file_path, "srt", mpl2_file_path, "txt", "mpl2")
        Utils.remove_trailing_newlines(new_mpl2_file_path, encoding)

    @staticmethod
    def mpl22srt(mpl2_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert MPL2 subtitles to SubRip subtitles.

        Arguments:
            mpl2_file_path {string} -- The path to the MPL2 file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        new_srt_file_path, encoding = Utils.__convert_subtitle(mpl2_file_path, "txt", srt_file_path, "srt", "srt")
        Utils.remove_trailing_newlines(new_srt_file_path, encoding)

    @staticmethod
    def srt2tmp(srt_file_path: str, tmp_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to TMP subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            tmp_file_path {string} -- The path to the TMP file.
        """

        new_tmp_file_path, encoding = Utils.__convert_subtitle(srt_file_path, "srt", tmp_file_path, "tmp", "tmp")
        Utils.remove_trailing_newlines(new_tmp_file_path, encoding)

    @staticmethod
    def tmp2srt(tmp_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert TMP subtitles to SubRip subtitles.

        Arguments:
            mpl2_file_path {string} -- The path to the TMP file.
            tmp_file_path {string} -- The path to the SubRip file.
        """

        new_srt_file_path, encoding = Utils.__convert_subtitle(tmp_file_path, "tmp", srt_file_path, "srt", "srt")
        Utils.remove_trailing_newlines(new_srt_file_path, encoding)

    @staticmethod
    def srt2sami(srt_file_path: str, sami_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to SAMI subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            sami_file_path {string} -- The path to the SAMI file.
        """

        file: Union[TextIO, BinaryIO]
        converter = CaptionConverter()
        encoding = Utils.detect_encoding(srt_file_path)
        with open(srt_file_path, "r", encoding=encoding) as file:
            converter.read(file.read(), SRTReader())
        if sami_file_path is None:
            sami_file_path = srt_file_path.replace(".srt", ".smi")
        with open(sami_file_path, "wb") as file:
            file.write(converter.write(SAMIWriter()).encode(encoding, errors="replace"))

    @staticmethod
    def sami2srt(sami_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert SAMI subtitles to SubRip subtitles.

        Arguments:
            sami_file_path {string} -- The path to the SAMI file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        file: Union[TextIO, BinaryIO]
        converter = CaptionConverter()
        encoding = Utils.detect_encoding(sami_file_path)
        with open(sami_file_path, "r", encoding=encoding) as file:
            converter.read(file.read(), SAMIReader())
        if srt_file_path is None:
            srt_file_path = sami_file_path.replace(".smi", ".srt")
        with open(srt_file_path, "wb") as file:
            file.write(converter.write(SRTWriter()).encode(encoding, errors="replace"))
        Utils.remove_trailing_newlines(srt_file_path, encoding)

    @staticmethod
    def stl2srt(stl_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert EBU-STL subtitles to SubRip subtitles.

        Arguments:
            stl_file_path {string} -- The path to the EBU-STL file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        encoding = Utils.detect_encoding(stl_file_path)
        stl = STL(stl_file_path, True)
        if srt_file_path is None:
            srt_file_path = stl_file_path.replace(".stl", ".srt")
        srt = SRT(srt_file_path)
        for sub in stl:
            (tci, tco, txt) = sub
            srt.write(tci, tco, txt, encoding)
        srt.file.close()
        stl.file.close()
        Utils.remove_trailing_newlines(srt_file_path, encoding)

    @staticmethod
    def srt2scc(srt_file_path: str, scc_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to SCC subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            scc_file_path {string} -- The path to the Scenarist Closed Captions file.
        """

        file: Union[TextIO, BinaryIO]
        converter = CaptionConverter()
        encoding = Utils.detect_encoding(srt_file_path)
        with open(srt_file_path, "r", encoding=encoding) as file:
            converter.read(file.read(), SRTReader())
        if scc_file_path is None:
            scc_file_path = srt_file_path.replace(".srt", ".scc")
        with open(scc_file_path, "wb") as file:
            file.write(converter.write(SCCWriter()).encode(encoding, errors="replace"))

    @staticmethod
    def scc2srt(scc_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert SCC subtitles to SubRip subtitles.

        Arguments:
            scc_file_path {string} -- The path to the Scenarist Closed Captions file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        file: Union[TextIO, BinaryIO]
        converter = CaptionConverter()
        encoding = Utils.detect_encoding(scc_file_path)
        with open(scc_file_path, "r", encoding=encoding) as file:
            converter.read(file.read(), SCCReader())
        if srt_file_path is None:
            srt_file_path = scc_file_path.replace(".scc", ".srt")
        with open(srt_file_path, "wb") as file:
            file.write(converter.write(SRTWriter()).encode(encoding, errors="replace"))
        Utils.remove_trailing_newlines(srt_file_path, encoding)

    @staticmethod
    def srt2sbv(srt_file_path: str, sbv_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to SubViewer subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            sbv_file_path {string} -- The path to the SubViewer file.
        """

        Caption.text = property(lambda self: self._text, Utils._set_text_patch)
        Reader.read = Utils._read_patch

        file: Union[TextIO, BinaryIO]
        encoding = Utils.detect_encoding(srt_file_path)
        with open(srt_file_path, "r", encoding=encoding) as file:
            srt_reader = SrtReader(file)
            srt_reader.encoding = encoding
            captions = srt_reader.read()
            for caption in captions:
                caption.encoding = encoding

        if sbv_file_path is None:
            sbv_file_path = srt_file_path.replace(".srt", ".sbv")

        with open(sbv_file_path, "w") as file:
            sbv_writer = SbvWriter(file, captions)
            sbv_writer.write()
            sbv_writer.close()

        Utils.remove_trailing_newlines(sbv_file_path, encoding)

    @staticmethod
    def sbv2srt(sbv_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert SubViewer subtitles to SubRip subtitles.

        Arguments:
            sbv_file_path {string} -- The path to the SubViewer file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        file: Union[TextIO, BinaryIO]
        encoding = Utils.detect_encoding(sbv_file_path)

        Caption.text = property(lambda self: self._text, Utils._set_text_patch)
        Reader.read = Utils._read_patch

        with open(sbv_file_path, "r", encoding=encoding) as file:
            sbv_reader = SbvReader(file)
            sbv_reader.encoding = encoding
            captions = sbv_reader.read()
            for caption in captions:
                caption.encoding = encoding

        if srt_file_path is None:
            srt_file_path = sbv_file_path.replace(".sbv", ".srt")

        with open(srt_file_path, "w") as file:
            srt_writer = SrtWriter(file, captions)
            srt_writer.write()
            srt_writer.close()

        Utils.remove_trailing_newlines(srt_file_path, encoding)

    @staticmethod
    def srt2ytt(srt_file_path: str, transcript_file_path: Optional[str] = None) -> None:
        """Convert SubRip subtitles to YouTube transcript subtitles.

        Arguments:
            srt_file_path {string} -- The path to the SubRip file.
            transcript_file_path {string} -- The path to the YouTube transcript file.
        """

        file: Union[TextIO, BinaryIO]
        encoding = Utils.detect_encoding(srt_file_path)

        Caption.text = property(lambda self: self._text, Utils._set_text_patch)
        Reader.read = Utils._read_patch
        TranscriptWriter.get_utime = Utils._get_utime_patch
        TranscriptWriter.format_time = Utils._format_time_patch
        TranscriptWriter.DOCUMENT_TPL = u"""<?xml version="1.0" encoding="{}" ?><transcript>%s</transcript>""".format(encoding)
        TranscriptWriter.CAPTION_TPL = u"""<text start="%(start)s" dur="%(duration)s">%(text)s</text>"""

        with open(srt_file_path, "r", encoding=encoding) as file:
            srt_reader = SrtReader(file)
            srt_reader.encoding = encoding
            captions = srt_reader.read()
            for caption in captions:
                caption.encoding = encoding

        if transcript_file_path is None:
            transcript_file_path = srt_file_path.replace(".srt", ".ytt")

        with open(transcript_file_path, "w") as file:
            transcript_writer = TranscriptWriter(file, captions)
            transcript_writer.write()
            transcript_writer.close()

        Utils.remove_trailing_newlines(transcript_file_path, encoding)

    @staticmethod
    def ytt2srt(transcript_file_path: str, srt_file_path: Optional[str] = None) -> None:
        """Convert YouTube transcript subtitles to SubRip subtitles.

        Arguments:
            transcript_file_path {string} -- The path to the YouTube transcript file.
            srt_file_path {string} -- The path to the SubRip file.
        """

        Caption.text = property(lambda self: self._text, Utils._set_text_patch)
        Reader.read = Utils._read_patch
        TranscriptReader.text_to_captions = Utils._text_to_captions_patch

        file: Union[TextIO, BinaryIO]
        encoding = Utils.detect_encoding(transcript_file_path)
        with open(transcript_file_path, "r", encoding=encoding) as file:
            transcript_reader = TranscriptReader(file)
            transcript_reader.encoding = encoding
            captions = transcript_reader.read()
            for caption in captions:
                caption.encoding = encoding

        if srt_file_path is None:
            srt_file_path = transcript_file_path.replace(".ytt", ".srt")

        with open(srt_file_path, "w") as file:
            srt_writer = SrtWriter(file, captions)
            srt_writer.write()
            srt_writer.close()

        Utils.remove_trailing_newlines(srt_file_path, encoding)

    @staticmethod
    def extract_teletext_as_subtitle(ts_file_path: str, page_num: int, output_file_path: str, timeout_secs: int = 30) -> None:
        """Extract DVB Teletext from MPEG transport stream files and convert them into the output format.

        Arguments:
            ts_file_path {string} -- The path to the Transport Stream file.
            page_num {int} -- The page number for the Teletext
            output_file_path {string} -- The path to the output file.
            timeout_secs {int} -- The timeout in seconds on extraction {default: 30}.
        """

        command = "{0} -y -fix_sub_duration -txt_page {1} -txt_format text -i '{2}' '{3}'".format(Utils.FFMPEG_BIN, page_num, ts_file_path, output_file_path)
        timeout_msg = "Timeout on extracting Teletext from transport stream: {} on page: {}".format(ts_file_path, page_num)
        error_msg = "Cannot extract Teletext from transport stream: {} on page: {}".format(ts_file_path, page_num)

        def _callback(returncode: int, std_err: str) -> None:
            if returncode != 0:
                raise TerminalException(
                    "Cannot extract Teletext from transport stream: {} on page: {} with error {}".format(
                        ts_file_path, page_num, std_err
                    )
                )
            Utils.remove_trailing_newlines(output_file_path, None)

        Utils._run_command(command, timeout_secs, timeout_msg, error_msg, _callback)

    @staticmethod
    def extract_matroska_subtitle(mkv_file_path: str, stream_index: int, output_file_path: str, timeout_secs: int = 30) -> None:
        """Extract subtitles from Matroska files and convert them into the output format.

        Arguments:
            mkv_file_path {string} -- The path to the Matroska file.
            stream_index {int} -- The index of the subtitle stream
            output_file_path {string} -- The path to the output file.
            timeout_secs {int} -- The timeout in seconds on extraction {default: 30}.
        """

        command = "{0} -y -i '{1}' -map 0:s:{2} '{3}'".format(Utils.FFMPEG_BIN, mkv_file_path, stream_index, output_file_path)
        timeout_msg = "Timeout on extracting the subtitle from file: {} with stream index: {}".format(mkv_file_path, stream_index)
        error_msg = "Cannot extract the subtitle from file: {} with stream index: {}".format(mkv_file_path, stream_index)

        def _callback(returncode: int, std_err: str) -> None:
            if returncode != 0:
                raise TerminalException(
                    "Cannot extract the subtitle from file: {} with stream index: {} with error {}".format(
                        mkv_file_path, stream_index, std_err
                    )
                )
            Utils.remove_trailing_newlines(output_file_path, None)
        Utils._run_command(command, timeout_secs, timeout_msg, error_msg, _callback)

    @staticmethod
    def suppress_lib_logs() -> None:
        import os
        import logging
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
        os.environ["TF_CPP_MIN_VLOG_LEVEL"] = "0"
        logging.getLogger("tensorflow").disabled = True

    @staticmethod
    def remove_trailing_newlines(source_file_path: str, encoding: Optional[str], target_file_path: Optional[str] = None) -> None:
        with open(source_file_path, "r", encoding=encoding) as file:
            content = file.read()
        if target_file_path is not None:
            with open(target_file_path, "w", encoding=encoding) as file:
                file.write(content.rstrip())
        else:
            with open(source_file_path, "w", encoding=encoding) as file:
                file.write(content.rstrip())

    @staticmethod
    def download_file(remote_file_url: str, local_file_path: str) -> None:
        r = requests.get(remote_file_url, verify=True, stream=True)
        r.raw.decode_content = True
        with open(local_file_path, "wb") as file:
            shutil.copyfileobj(r.raw, file)

    @staticmethod
    def contains_embedded_subtitles(video_file_path: str, timeout_secs: int = 30) -> bool:
        """Detect if the input video contains embedded subtitles.

        Arguments:
            video_file_path {string} -- The path to the video file.
            timeout_secs {int} -- The timeout in seconds on extraction {default: 30}.

        Returns:
            bool -- True if the video contains embedded subtitles or False otherwise.
        """

        command = "{0} -y -i '{1}' -c copy -map 0:s -f null - -v 0 -hide_banner".format(Utils.FFMPEG_BIN, video_file_path)
        timeout_msg = "Timeout on detecting embedded subtitles from file: {}".format(video_file_path)
        error_msg = "Embedded subtitle detection failed for file: {}".format(video_file_path)

        def _callback(returncode: int, std_err: str) -> bool:
            return returncode == 0
        return Utils._run_command(command, timeout_secs, timeout_msg, error_msg, _callback)

    @staticmethod
    def detect_encoding(subtitle_file_path: str) -> str:
        """Detect the encoding of the subtitle file.

        Arguments:
            subtitle_file_path {string} -- The path to the subtitle file.

        Returns:
            string -- The string represent the encoding
        """

        with open(subtitle_file_path, "rb") as file:
            # raw = b"".join([file.readline() for _ in range(10)])
            # Sampling with 10 lines did not work well enough for large subtitle files
            # and hence this less memory-efficient solution:
            raw = b"".join(file.readlines())

        detected = cchardet.detect(raw)
        detected = detected or {}
        return detected["encoding"] if "encoding" in detected else None

    @staticmethod
    def get_file_root_and_extension(file_path: str) -> Tuple[str, str]:
        """Get the root path and the extension of the input file path.

        Returns:
            tuple -- the root path and the extension of the input file path.
        """

        parts = os.path.abspath(file_path).split(os.extsep, 1)
        return parts[0], parts[1]

    @staticmethod
    def get_stretch_language_codes() -> List[str]:
        """Get language codes used by stretch.

        Returns:
            list -- A list of language codes derived from ISO 639-3.
        """
        return Language.ALLOWED_VALUES

    @staticmethod
    def get_misc_language_codes() -> List[str]:
        """Get all known language codes.

        Returns:
            list -- A list of all known language codes.
        """
        return Language.ALLOWED_VALUES + \
            ['CELTIC', 'NORTH_EU', 'NORWAY', 'ROMANCE', 'SAMI', 'SCANDINAVIA', 'aav', 'aed', 'afa', 'alv', 'art', 'ase',
             'bat', 'bcl', 'bem', 'ber', 'bnt', 'bzs', 'cau', 'ccs', 'ceb', 'cel', 'chk', 'cpf', 'cpp', 'crs', 'csg',
             'csn', 'cus', 'dra', 'efi', 'en_el_es_fi', 'euq', 'fi_nb_no_nn_ru_sv_en', 'fiu', 'fse', 'gaa', 'gem',
             'gil', 'gmq', 'gmw', 'grk', 'guw', 'hil', 'iir', 'ilo', 'inc', 'ine', 'iso', 'itc', 'jap', 'kab', 'kqn',
             'kwn', 'kwy', 'loz', 'lua', 'lue', 'lun', 'luo', 'lus', 'map', 'mfe', 'mfs', 'mkh', 'mos', 'mul', 'nic',
             'niu', 'nso', 'nyk', 'pag', 'phi', 'pis', 'pon', 'poz', 'pqe', 'pqw', 'prl', 'rnd', 'roa', 'run', 'sal',
             'sem', 'sit', 'sla', 'srn', 'ssp', 'swc', 'taw', 'tdt', 'tiv', 'tll', 'toi', 'tpi', 'trk', 'tum', 'tut',
             'tvl', 'tzo', 'umb', 'urj', 'vsl', 'wal', 'war', 'wls', 'yap', 'yua', 'zai', 'zle', 'zls', 'zlw', 'zne']

    @staticmethod
    def get_language_table() -> List[str]:
        """Get all known language codes and their human-readable versions.

        Returns:
            list -- A list of all known language codes and their human-readable versions.
        """
        return list(map(lambda line: line.replace("\t", "  "), Language.CODE_TO_HUMAN_LIST)) + \
            ['CELTIC', 'NORTH_EU', 'NORWAY', 'ROMANCE', 'SAMI', 'SCANDINAVIA', 'aav', 'aed', 'afa', 'alv', 'art', 'ase',
             'bat', 'bcl', 'bem', 'ber', 'bnt', 'bzs', 'cau', 'ccs', 'ceb', 'cel', 'chk', 'cpf', 'cpp', 'crs', 'csg',
             'csn', 'cus', 'dra', 'efi', 'en_el_es_fi', 'euq', 'fi_nb_no_nn_ru_sv_en', 'fiu', 'fse', 'gaa', 'gem',
             'gil', 'gmq', 'gmw', 'grk', 'guw', 'hil', 'iir', 'ilo', 'inc', 'ine', 'iso', 'itc', 'jap', 'kab', 'kqn',
             'kwn', 'kwy', 'loz', 'lua', 'lue', 'lun', 'luo', 'lus', 'map', 'mfe', 'mfs', 'mkh', 'mos', 'mul', 'nic',
             'niu', 'nso', 'nyk', 'pag', 'phi', 'pis', 'pon', 'poz', 'pqe', 'pqw', 'prl', 'rnd', 'roa', 'run', 'sal',
             'sem', 'sit', 'sla', 'srn', 'ssp', 'swc', 'taw', 'tdt', 'tiv', 'tll', 'toi', 'tpi', 'trk', 'tum', 'tut',
             'tvl', 'tzo', 'umb', 'urj', 'vsl', 'wal', 'war', 'wls', 'yap', 'yua', 'zai', 'zle', 'zls', 'zlw', 'zne']

    @staticmethod
    def __convert_subtitle(source_file_path: str, source_ext: str, target_file_path: Optional[str], target_ext: str, format: str, frame_rate: Optional[float] = None) -> Tuple[str, str]:
        encoding = Utils.detect_encoding(source_file_path)
        subs = pysubs2.load(source_file_path, encoding=encoding)
        new_target_file_path = source_file_path.replace(".%s" % source_ext, ".%s" % target_ext) if target_file_path is None else target_file_path
        if frame_rate is None:
            subs.save(new_target_file_path, encoding=encoding, format_=format)
        else:
            subs.save(new_target_file_path, encoding=encoding, format_=format, fps=frame_rate)
        return new_target_file_path, encoding

    @staticmethod
    def _run_command(command: str, timeout_secs: int, timeout_msg: str, error_msg: str, callback: Callable[[int, str], Any]) -> Any:
        with subprocess.Popen(
                shlex.split(command),
                shell=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                close_fds=True,
                universal_newlines=True,
                bufsize=1,
        ) as process:
            try:
                _, std_err = process.communicate(timeout=timeout_secs)
                return callback(process.returncode, std_err)
            except subprocess.TimeoutExpired as te:
                process.kill()
                raise TerminalException(timeout_msg) from te
            except Exception as e:
                process.kill()
                if isinstance(e, TerminalException):
                    raise e
                else:
                    raise TerminalException(error_msg) from e
            finally:
                os.system("stty sane")

    @staticmethod
    def _set_text_patch(self_patched: Any, value: str) -> None:
        self_patched._text = value

    @staticmethod
    def _read_patch(self_patched: Any) -> List:
        self_patched.rawcontent = self_patched.fileobject.read()
        self_patched.text_to_captions()
        return self_patched.captions

    @staticmethod
    def _text_to_captions_patch(self_patched: Any) -> List:
        soup = BeautifulSoup(self_patched.rawcontent, features="lxml")
        texts = soup.find_all('text')
        for text in texts:
            caption = Caption()
            caption.start = self_patched.get_start(text)
            caption.duration = self_patched.get_duration(text)
            caption.text = text.text
            self_patched.add_caption(caption)

        return self_patched.captions

    @staticmethod
    def _get_utime_patch(_: Any, dt: datetime) -> str:
        start = dt
        start_seconds = 3600 * start.hour + 60 * start.minute + start.second
        start_milliseconds = start.microsecond // 1000

        if start_milliseconds:
            ustart = u"%s.%s" % (start_seconds, start_milliseconds)
        else:
            ustart = u"%s" % start_seconds

        return ustart

    @staticmethod
    def _format_time_patch(self_patched: Any, caption: Any) -> Dict:
        return {
            "start": self_patched.get_utime(caption.start),
            "end": self_patched.get_utime(caption.end),
            "duration": caption.duration.total_seconds()
        }



import hyperopt
import numpy as np
from hyperopt import hp
from hyperopt.pyll.base import scope
from typing import List, Dict
from .trainer import Trainer
from .embedder import FeatureEmbedder
from .hyperparameters import Hyperparameters
from .network import Network


class HyperParameterTuner(object):
    """Hyperparameter tuning using the Tree of Parzen Estimators algorithm"""

    SEARCH_SPACE = {
        "learning_rate": hp.loguniform("learning_rate", np.log(0.00001), np.log(0.1)),
        "front_hidden_size": hp.choice("front_hidden_size", [[16], [32], [64]]),
        "back_hidden_size": hp.choice("back_hidden_size", [[16, 8], [32, 16]]),
        "batch_size": scope.int(hp.qloguniform("batch_size", np.log(8), np.log(4096), 1)),
        "optimizer": hp.choice("optimizer", Hyperparameters.OPTIMIZERS),
        "dropout": hp.loguniform("dropout", np.log(0.1), np.log(0.3)),
        "validation_split": hp.loguniform("validation_split", np.log(0.2), np.log(0.4)),
    }

    def __init__(self,
                 av_file_paths: List[str],
                 subtitle_file_paths: List[str],
                 training_dump_dir: str,
                 num_of_trials: int = 5,
                 tuning_epochs: int = 5,
                 network_type: str = Network.LSTM,
                 **kwargs) -> None:
        """Hyperparameter tuner initialiser

        Arguments:
            av_file_paths {list} -- A list of paths to the input audio/video files.
            subtitle_file_paths {list} -- A list of paths to the subtitle files.
            training_dump_dir {string} --  The directory of the training data dump file.

        Keyword Arguments:
            num_of_trials {int} -- The number of trials for tuning (default: {5}).
            tuning_epochs {int} -- The number of training epochs for each trial (default: {5}).
            network_type {string} -- The type of the network (default: {"lstm"}, range: ["lstm", "bi_lstm", "conv_1d"]).
        """

        assert network_type in Network.TYPES, "Supported network type values: %s" % Network.TYPES
        hyperparameters = Hyperparameters()
        hyperparameters.network_type = network_type
        self.__hyperparameters = hyperparameters

        self.__trainer = Trainer(FeatureEmbedder(**kwargs))
        self.__av_file_paths = av_file_paths
        self.__subtitle_file_paths = subtitle_file_paths
        self.__training_dump_dir = training_dump_dir
        self.__num_of_trials = num_of_trials
        self.__tuning_epochs = tuning_epochs
        self.__original_epochs = self.__hyperparameters.epochs

    @property
    def hyperparameters(self) -> Hyperparameters:
        self.__hyperparameters.epochs = self.__original_epochs
        return self.__hyperparameters.clone()

    def tune_hyperparameters(self) -> None:
        """Tune the hyperparameters"""

        trials = hyperopt.Trials()
        minimised = hyperopt.fmin(fn=self.__get_val_loss,
                                  space=self.SEARCH_SPACE,
                                  trials=trials,
                                  algo=hyperopt.tpe.suggest,
                                  max_evals=self.__num_of_trials,
                                  show_progressbar=False)
        turned_hps = hyperopt.space_eval(self.SEARCH_SPACE, minimised)
        for key, value in turned_hps.items():
            if key == "front_hidden_size":
                self.__hyperparameters.front_hidden_size = list(value)
            elif key == "back_hidden_size":
                self.__hyperparameters.back_hidden_size = list(value)
            else:
                setattr(self.__hyperparameters, key, value)

    def __get_val_loss(self, params: Dict) -> Dict:
        for key, value in params.items():
            if key == "front_hidden_size":
                self.__hyperparameters.front_hidden_size = list(value)
            elif key == "back_hidden_size":
                self.__hyperparameters.back_hidden_size = list(value)
            else:
                setattr(self.__hyperparameters, key, value)

        self.__hyperparameters.epochs = self.__tuning_epochs
        val_loss, _ = self.__trainer.pre_train(self.__av_file_paths,
                                               self.__subtitle_file_paths,
                                               self.__training_dump_dir,
                                               self.__hyperparameters)
        if not val_loss:
            raise ValueError("Cannot get training loss during hyperparameter tuning")

        return {"loss": sum(val_loss) / len(val_loss), "status": hyperopt.STATUS_OK}



import subprocess
import os
import threading
import traceback
import tempfile
import shutil
import atexit
import signal
import shlex

from typing import Optional, Tuple, List
from copy import deepcopy
from pysrt import SubRipFile, SubRipItem
from decimal import Decimal
from .embedder import FeatureEmbedder
from .exception import TerminalException
from .exception import NoFrameRateException
from .logger import Logger

TEMP_DIR_PATH = tempfile.mkdtemp()


def clear_temp(*_):
    if os.path.isdir(TEMP_DIR_PATH):
        shutil.rmtree(TEMP_DIR_PATH)


class MediaHelper(object):
    """ Utility for processing media assets including audio, video and
    subtitle files.
    """

    FFMPEG_BIN = os.getenv("FFMPEG_PATH") or os.getenv("ffmpeg_path") or "ffmpeg"

    AUDIO_FILE_EXTENSION = [".wav", ".aac"]

    __LOGGER = Logger().get_logger(__name__)
    __MIN_SECS_PER_WORD = 0.414  # 60 secs / 145 wpm
    __MIN_GAP_IN_SECS = (
        1  # minimum gap in seconds between consecutive subtitle during segmentation
    )
    __CMD_TIME_OUT = 180  # time out for subprocess

    atexit.register(clear_temp)
    signal.signal(signal.SIGTERM, clear_temp)

    @staticmethod
    def extract_audio(video_file_path, decompress: bool = False, freq: int = 16000) -> str:
        """Extract audio track from the video file and save it to a WAV file.

        Arguments:
            video_file_path {string} -- The input video file path.
        Keyword Arguments:
            decompress {bool} -- Extract WAV if True otherwise extract AAC (default: {False}).
            freq {int} -- The audio sample frequency (default: {16000}).
        Returns:
            string -- The file path of the extracted audio.
        """

        basename = os.path.basename(video_file_path)

        # Using WAV for training or prediction is faster than using AAC.
        # However the former will result in larger temporary audio files saved on the disk.
        if decompress:
            assert freq is not None, "Frequency is needed for decompression"
            audio_file_path = "{0}/{1}{2}".format(
                TEMP_DIR_PATH, basename, MediaHelper.AUDIO_FILE_EXTENSION[0]
            )
        else:
            audio_file_path = "{0}/{1}{2}".format(
                TEMP_DIR_PATH, basename, MediaHelper.AUDIO_FILE_EXTENSION[1]
            )

        command = (
            "{0} -y -xerror -i '{1}' -ac 2 -ar {2} -vn '{3}'".format(
                MediaHelper.FFMPEG_BIN, video_file_path, freq, audio_file_path
            )
            if decompress
            else "{0} -y -xerror -i '{1}' -vn -acodec copy '{2}'".format(
                MediaHelper.FFMPEG_BIN, video_file_path, audio_file_path
            )
        )
        with subprocess.Popen(
            shlex.split(command),
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            universal_newlines=True,
            bufsize=1,
        ) as process:
            try:
                MediaHelper.__LOGGER.debug("[{}-{}] Running: {}".format(threading.current_thread().name, process.pid, command))
                _, std_err = process.communicate(timeout=MediaHelper.__CMD_TIME_OUT)
                MediaHelper.__LOGGER.debug("[{}-{}] {}".format(threading.current_thread().name, process.pid, std_err))
                if process.returncode != 0:
                    MediaHelper.__LOGGER.error("[{}-{}] Cannot extract audio from video: {}\n{}"
                                               .format(threading.current_thread().name, process.pid, video_file_path, std_err))
                    raise TerminalException(
                        "Cannot extract audio from video: {}".format(video_file_path)
                    )
                MediaHelper.__LOGGER.info(
                    "[{}-{}] Extracted audio file: {}".format(threading.current_thread().name, process.pid,
                                                              audio_file_path))
                return audio_file_path
            except subprocess.TimeoutExpired as te:
                MediaHelper.__LOGGER.error("Timeout on extracting audio from video: {}".format(video_file_path))
                if os.path.exists(audio_file_path):
                    os.remove(audio_file_path)
                raise TerminalException(
                    "Timeout on extracting audio from video: {}".format(video_file_path)
                ) from te
            except Exception as e:
                if os.path.exists(audio_file_path):
                    os.remove(audio_file_path)
                if isinstance(e, TerminalException):
                    raise e
                else:
                    raise TerminalException(
                        "Cannot extract audio from video: {}".format(video_file_path)
                    ) from e
            except KeyboardInterrupt:
                MediaHelper.__LOGGER.error(
                    "[{}-{}] Extracting audio from video {} interrupted".format(
                        threading.current_thread().name, process.pid, video_file_path
                    )
                )
                if os.path.exists(audio_file_path):
                    os.remove(audio_file_path)
                process.send_signal(signal.SIGINT)
                raise TerminalException(
                    "Extracting audio from video {} interrupted".format(video_file_path)
                )
            finally:
                process.kill()
                os.system("stty sane")

    @staticmethod
    def get_duration_in_seconds(start: Optional[str], end: Optional[str]) -> Optional[float]:
        """Get the duration in seconds between a start time and an end time.

        Arguments:
            start {string} -- The start time (e.g., 00:00:00,750).
            end {string} -- The end time (e.g., 00:00:10,230).

        Returns:
            float -- The duration in seconds.
        """

        if start is None:
            start = "00:00:00,000"
        if end is None:
            return None
        start = start.replace(",", ".")
        end = end.replace(",", ".")
        start_h, start_m, start_s = map(Decimal, start.split(":"))
        end_h, end_m, end_s = map(Decimal, end.split(":"))
        return float(
            (end_h * 3600 + end_m * 60 + end_s)
            - (start_h * 3600 + start_m * 60 + start_s)
        )

    @staticmethod
    def extract_audio_from_start_to_end(audio_file_path: str, start: str, end: Optional[str] = None) -> Tuple[str, Optional[float]]:
        """Extract audio based on the start time and the end time and save it to a temporary file.

        Arguments:
            audio_file_path {string} -- The path of the audio file.
            start {string} -- The start time (e.g., 00:00:00,750).

        Keyword Arguments:
            end {string} -- The end time (e.g., 00:00:10,230) (default: {None}).

        Returns:
            tuple -- The file path to the extracted audio and its duration.
        """

        segment_duration = MediaHelper.get_duration_in_seconds(start, end)
        basename = os.path.basename(audio_file_path)
        filename, extension = os.path.splitext(basename)
        start = start.replace(",", ".")
        if end is not None:
            end = end.replace(",", ".")
        segment_path = "{0}/{1}_{2}_{3}{4}".format(TEMP_DIR_PATH, filename, str(start), str(end), extension)

        if end is not None:
            duration = MediaHelper.get_duration_in_seconds(start, end)
            command = "{0} -y -xerror -i '{1}' -ss {2} -t {3} -acodec copy '{4}'".format(
                MediaHelper.FFMPEG_BIN, audio_file_path, start, duration, segment_path
            )
        else:
            command = "{0} -y -xerror -i '{1}' -ss {2} -acodec copy '{3}'".format(
                MediaHelper.FFMPEG_BIN, audio_file_path, start, segment_path
            )
        with subprocess.Popen(
            shlex.split(command),
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            bufsize=1,
        ) as process:
            MediaHelper.__LOGGER.debug("[{}-{}] Running: {}".format(threading.current_thread().name, process.pid, command))
            try:
                _, std_err = process.communicate(timeout=MediaHelper.__CMD_TIME_OUT)
                MediaHelper.__LOGGER.debug("[{}-{}] {}".format(threading.current_thread().name, process.pid, std_err))
                if process.returncode != 0:
                    MediaHelper.__LOGGER.error("[{}-{}] Cannot clip audio: {} Return Code: {}\n{}"
                                               .format(threading.current_thread().name, process.pid, audio_file_path, process.returncode, std_err))
                    raise TerminalException(
                        "Cannot clip audio: {} Return Code: {}".format(audio_file_path, process.returncode)
                    )
                MediaHelper.__LOGGER.info(
                    "[{}-{}] Extracted audio segment: {}".format(threading.current_thread().name, process.pid,
                                                                 segment_path))
                return segment_path, segment_duration
            except subprocess.TimeoutExpired as e:
                MediaHelper.__LOGGER.error(
                    "[{}-{}] Extracting {} timed out: {}\n{}".format(
                        threading.current_thread().name, process.pid, segment_path, str(e), "\n".join(traceback.format_stack())
                    )
                )
                traceback.print_tb(e.__traceback__)
                if os.path.exists(segment_path):
                    os.remove(segment_path)
                raise TerminalException(
                    "Timeout on extracting audio from audio: {} after {} seconds".format(audio_file_path, MediaHelper.__CMD_TIME_OUT)
                ) from e
            except Exception as e:
                MediaHelper.__LOGGER.error(
                    "[{}-{}] Extracting {} failed: {}\n{}".format(
                        threading.current_thread().name, process.pid, segment_path, str(e), "\n".join(traceback.format_stack())
                    )
                )
                traceback.print_tb(e.__traceback__)
                if os.path.exists(segment_path):
                    os.remove(segment_path)
                if isinstance(e, TerminalException):
                    raise e
                else:
                    raise TerminalException(
                        "Cannot clip audio: {}".format(audio_file_path)
                    ) from e
            except KeyboardInterrupt:
                MediaHelper.__LOGGER.error(
                    "[{}-{}] Extracting with start and end from {} interrupted".format(
                        threading.current_thread().name, process.pid, segment_path
                    )
                )
                if os.path.exists(segment_path):
                    os.remove(segment_path)
                process.send_signal(signal.SIGINT)
                raise TerminalException("Extracting with start and end from {} interrupted".format(segment_path))
            finally:
                process.kill()
                os.system("stty sane")

    @staticmethod
    def get_audio_segment_starts_and_ends(subs: List[SubRipItem]) -> Tuple[List[str], List[str], List[SubRipFile]]:
        """Group subtitle cues into larger segments in terms of silence gaps.

        Arguments:
            subs {list} -- A list of SupRip cues.

        Returns:
            tuple -- A list of start times, a list of end times and a list of grouped SubRip files.
        """

        local_subs = MediaHelper.__preprocess_subs(subs)

        segment_starts = []
        segment_ends = []
        combined = []
        new_subs = []
        current_start = str(local_subs[0].start)

        for i in range(len(local_subs)):
            if i == len(local_subs) - 1:
                combined.append(local_subs[i])
                segment_starts.append(current_start)
                segment_ends.append(str(local_subs[i].end))
                new_subs.append(SubRipFile(combined))
                del combined[:]
            else:
                # Do not segment when the subtitle is too short
                duration = FeatureEmbedder.time_to_sec(
                    local_subs[i].end
                ) - FeatureEmbedder.time_to_sec(local_subs[i].start)
                if duration < MediaHelper.__MIN_SECS_PER_WORD:
                    combined.append(local_subs[i])
                    continue
                # Do not segment consecutive subtitles having little or no gap.
                gap = FeatureEmbedder.time_to_sec(
                    local_subs[i + 1].start
                ) - FeatureEmbedder.time_to_sec(local_subs[i].end)
                if (
                    local_subs[i].end == local_subs[i + 1].start
                    or gap < MediaHelper.__MIN_GAP_IN_SECS
                ):
                    combined.append(local_subs[i])
                    continue
                combined.append(local_subs[i])
                # The start time is set to last cue's end time
                segment_starts.append(current_start)
                # The end time cannot be set to next cue's start time due to possible overlay
                segment_ends.append(str(local_subs[i].end))
                current_start = str(local_subs[i].end)
                new_subs.append(SubRipFile(combined))
                del combined[:]
        return segment_starts, segment_ends, new_subs

    @staticmethod
    def get_frame_rate(file_path: str) -> float:
        """Extract the video frame rate. Will return 25 when input is audio

        Arguments:
            file_path {string} -- The input audiovisual file path.
        Returns:
            float -- The frame rate
        """

        with subprocess.Popen(
                shlex.split("{0} -i '{1}' -t 00:00:10 -f null /dev/null".format(MediaHelper.FFMPEG_BIN, file_path)),
                shell=False,
                stderr=subprocess.PIPE,
                close_fds=True,
                universal_newlines=True,
                bufsize=1,
        ) as proc:
            with subprocess.Popen(
                    ['grep', '-Eo', r"[0-9]{1,3}(\.[0-9]{1,3})?\sfps,"],
                    shell=False,
                    stdin=proc.stderr,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    close_fds=True,
                    universal_newlines=True,
                    bufsize=1,
            ) as process:
                try:
                    std_out, std_err = process.communicate(timeout=MediaHelper.__CMD_TIME_OUT)
                    if process.returncode != 0:
                        MediaHelper.__LOGGER.warning("[{}-{}] Cannot extract the frame rate from video: {}\n{}".format(threading.current_thread().name, process.pid, file_path, std_err))
                        raise NoFrameRateException(
                            "Cannot extract the frame rate from video: {}".format(file_path)
                        )
                    fps = float(std_out.split(" ")[0])
                    # ffmpeg uses two decimal places so be this hack
                    fps = fps if fps != 23.98 else 23.976
                    MediaHelper.__LOGGER.info("[{}-{}] Extracted frame rate: {} fps".format(threading.current_thread().name, process.pid, fps))
                    return fps
                except subprocess.TimeoutExpired as te:
                    raise NoFrameRateException(
                        "Timeout on extracting the frame rate from video: {}".format(file_path)
                    ) from te
                except Exception as e:
                    if isinstance(e, TerminalException):
                        raise e
                    else:
                        raise NoFrameRateException(
                            "Cannot extract the frame rate from video: {}".format(file_path)
                        ) from e
                except KeyboardInterrupt:
                    MediaHelper.__LOGGER.error(
                        "[{}-{}] Extracting frame rate from video {} interrupted".format(
                            threading.current_thread().name, process.pid, file_path
                        )
                    )
                    process.send_signal(signal.SIGINT)
                    proc.send_signal(signal.SIGINT)
                    raise TerminalException("Extracting frame rate from video {} interrupted".format(file_path))
                finally:
                    process.kill()
                    proc.kill()
                    os.system("stty sane")

    @staticmethod
    def refragment_with_min_duration(subs: List[SubRipItem], minimum_segment_duration: float) -> List[SubRipItem]:
        """Re-fragment a list of subtitle cues into new cues each of spans a minimum duration

        Arguments:
            subs {list} -- A list of SupRip cues.
            minimum_segment_duration {float} -- The minimum duration in seconds for each output subtitle cue.
        Returns:
            list -- A list of new SupRip cues after fragmentation.
        """
        new_segment = []
        new_segment_index = 0
        new_segment_duration = 0.0
        new_segment_text = ""
        new_subs = []
        for sub in subs:
            if minimum_segment_duration > new_segment_duration:
                new_segment.append(sub)
                new_segment_duration += MediaHelper.get_duration_in_seconds(str(sub.start), str(sub.end)) or 0.0
                new_segment_text += "{}\n".format(sub.text)
            else:
                concatenated_item = SubRipItem(new_segment_index, new_segment[0].start, new_segment[-1].end,
                                               new_segment_text, new_segment[0].position)
                new_subs.append(concatenated_item)
                new_segment_index += 1
                new_segment = [sub]
                new_segment_duration = MediaHelper.get_duration_in_seconds(str(sub.start), str(sub.end)) or 0.0
                new_segment_text = "{}\n".format(sub.text)
        if new_segment:
            concatenated_item = SubRipItem(new_segment_index, new_segment[0].start, new_segment[-1].end,
                                           new_segment_text, new_segment[0].position)
            new_subs.append(concatenated_item)
        return new_subs

    @staticmethod
    def __preprocess_subs(subs: List[SubRipItem]) -> List[SubRipItem]:
        local_subs = deepcopy(subs)

        # Preprocess overlapping subtitles
        for i in range(len(local_subs)):
            if i != 0 and local_subs[i].start < local_subs[i - 1].end:
                MediaHelper.__LOGGER.warning("Found overlapping subtitle cues and the earlier one's duration will be shortened.")
                local_subs[i - 1].end = local_subs[i].start

        return local_subs



#!/usr/bin/env python
"""
usage: subaligner_batch [-h] [-m {single,dual}] [-vd VIDEO_DIRECTORY] [-sd SUBTITLE_DIRECTORY] [-l MAX_LOGLOSS] [-so]
                        [-sil {afr,amh,ara,arg,asm,aze,ben,bos,bul,cat,ces,cmn,cym,dan,deu,ell,eng,epo,est,eus,fas,fin,fra,gla,gle,glg,grc,grn,guj,heb,hin,hrv,hun,hye,ina,ind,isl,ita,jbo,jpn,kal,kan,kat,kir,kor,kur,lat,lav,lfn,lit,mal,mar,mkd,mlt,msa,mya,nah,nep,nld,nor,ori,orm,pan,pap,pol,por,ron,rus,sin,slk,slv,spa,sqi,srp,swa,swe,tam,tat,tel,tha,tsn,tur,ukr,urd,vie,yue,zho}]
                        [-fos] [-tod TRAINING_OUTPUT_DIRECTORY] [-od OUTPUT_DIRECTORY] [-t TRANSLATE] [-lgs] [-d] [-q] [-ver]

Batch align multiple subtitle files and audiovisual files (v0.1.4)

Subtitle files and their companion audiovisual files need to be stored in two separate directories.
Each file pair needs to share the same base filename, the part before the extension.

optional arguments:
  -h, --help            show this help message and exit
  -vd VIDEO_DIRECTORY, --video_directory VIDEO_DIRECTORY
                        Path to the video directory
  -sd SUBTITLE_DIRECTORY, --subtitle_directory SUBTITLE_DIRECTORY
                        Path to the subtitle directory
  -l MAX_LOGLOSS, --max_logloss MAX_LOGLOSS
                        Max global log loss for alignment
  -so, --stretch_off    Switch off stretch on non-English speech and subtitles)
  -sil {afr,amh,ara,arg,asm,aze,ben,bos,bul,cat,ces,cmn,cym,dan,deu,ell,eng,epo,est,eus,fas,fin,fra,gla,gle,glg,grc,grn,guj,heb,hin,hrv,hun,hye,ina,ind,isl,ita,jbo,jpn,kal,kan,kat,kir,kor,kur,lat,lav,lfn,lit,mal,mar,mkd,mlt,msa,mya,nah,nep,nld,nor,ori,orm,pan,pap,pol,por,ron,rus,sin,slk,slv,spa,sqi,srp,swa,swe,tam,tat,tel,tha,tsn,tur,ukr,urd,vie,yue,zho}, --stretch_in_language {afr,amh,ara,arg,asm,aze,ben,bos,bul,cat,ces,cmn,cym,dan,deu,ell,eng,epo,est,eus,fas,fin,fra,gla,gle,glg,grc,grn,guj,heb,hin,hrv,hun,hye,ina,ind,isl,ita,jbo,jpn,kal,kan,kat,kir,kor,kur,lat,lav,lfn,lit,mal,mar,mkd,mlt,msa,mya,nah,nep,nld,nor,ori,orm,pan,pap,pol,por,ron,rus,sin,slk,slv,spa,sqi,srp,swa,swe,tam,tat,tel,tha,tsn,tur,ukr,urd,vie,yue,zho}
                        Stretch the subtitle with the supported ISO 639-3 language code [https://en.wikipedia.org/wiki/List_of_ISO_639-3_codes].
                        NB: This will be ignored if either -so or --stretch_off is present
  -fos, --exit_segfail  Exit on any segment alignment failures
  -tod TRAINING_OUTPUT_DIRECTORY, --training_output_directory TRAINING_OUTPUT_DIRECTORY
                        Path to the output directory containing training results
  -od OUTPUT_DIRECTORY, --output_directory OUTPUT_DIRECTORY
                        Path to the output subtitle directory
  -t TRANSLATE, --translate TRANSLATE
                        Source and target ISO 639-3 language codes separated by a comma (e.g., eng,zho)
  -lgs, --languages     Print out language codes used for stretch and translation
  -d, --debug           Print out debugging information
  -q, --quiet           Switch off logging information
  -ver, --version       show program's version number and exit

required arguments:
  -m {single,dual}, --mode {single,dual}
                        Alignment mode: either single or dual
"""

import argparse
import sys
import traceback
import os


def main():
    if sys.version_info.major != 3:
        print("Cannot find Python 3")
        sys.exit(20)
    try:
        import subaligner
    except ModuleNotFoundError:
        print("Subaligner is not installed")
        sys.exit(20)

    from subaligner._version import __version__
    parser = argparse.ArgumentParser(description="""Batch align multiple subtitle files and audiovisual files (v%s)\n
Subtitle files and their companion audiovisual files need to be stored in two separate directories.
Each file pair needs to share the same base filename, the part before the extension.""" % __version__, formatter_class=argparse.RawTextHelpFormatter)
    required_args = parser.add_argument_group("required arguments")
    required_args.add_argument(
        "-m",
        "--mode",
        type=str,
        default="",
        choices=["single", "dual"],
        help="Alignment mode: either single or dual",
    )
    parser.add_argument(
        "-vd",
        "--video_directory",
        type=str,
        default="",
        help="Path to the video directory",
    )
    parser.add_argument(
        "-sd",
        "--subtitle_directory",
        type=str,
        default="",
        help="Path to the subtitle directory",
    )
    parser.add_argument(
        "-l",
        "--max_logloss",
        type=float,
        default=float("inf"),
        help="Max global log loss for alignment",
    )
    parser.add_argument(
        "-so",
        "--stretch_off",
        action="store_true",
        help="Switch off stretch on non-English speech and subtitles)",
    )
    from subaligner.utils import Utils
    parser.add_argument(
        "-sil",
        "--stretch_in_language",
        type=str,
        choices=Utils.get_stretch_language_codes(),
        default="eng",
        help="Stretch the subtitle with the supported ISO 639-3 language code [https://en.wikipedia.org/wiki/List_of_ISO_639-3_codes].\nNB: This will be ignored if either -so or --stretch_off is present",
    )
    parser.add_argument(
        "-fos",
        "--exit_segfail",
        action="store_true",
        help="Exit on any segment alignment failures",
    )
    parser.add_argument(
        "-tod",
        "--training_output_directory",
        type=str,
        default=os.path.abspath(os.path.dirname(subaligner.__file__)),
        help="Path to the output directory containing training results",
    )
    parser.add_argument(
        "-od",
        "--output_directory",
        type=str,
        default="",
        help="Path to the output subtitle directory",
    )
    parser.add_argument(
        "-t",
        "--translate",
        type=str,
        help="Source and target ISO 639-3 language codes separated by a comma (e.g., eng,zho)",
    )
    parser.add_argument("-lgs", "--languages", action="store_true",
                        help="Print out language codes used for stretch and translation")
    parser.add_argument("-d", "--debug", action="store_true",
                        help="Print out debugging information")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="Switch off logging information")
    parser.add_argument("-ver", "--version", action="version", version=__version__)
    FLAGS, unparsed = parser.parse_known_args()

    if FLAGS.languages:
        print("\n".join(Utils.get_language_table()))
        sys.exit(0)
    if FLAGS.mode == "":
        print("--mode was not passed in")
        parser.print_usage()
        sys.exit(21)
    if FLAGS.video_directory == "":
        print("--video_directory was not passed in")
        parser.print_usage()
        sys.exit(21)
    if FLAGS.subtitle_directory == "":
        print("--subtitle_directory was not passed in")
        parser.print_usage()
        sys.exit(21)
    if FLAGS.output_directory == "":
        print("--output_directory was not passed in")
        parser.print_usage()
        sys.exit(21)
    if os.path.abspath(FLAGS.subtitle_directory) == os.path.abspath(FLAGS.output_directory):
        print("The output directory cannot be set to the same as the input subtitle directory")
        parser.print_usage()
        sys.exit(21)

    video_file_paths = [os.path.abspath(os.path.join(path, p)) for path, _, files in
                        os.walk(FLAGS.video_directory) for p in files if not p.startswith(".")]
    subtitle_file_paths = [os.path.abspath(os.path.join(path, p)) for path, _, files in
                           os.walk(FLAGS.subtitle_directory) for p in files if not p.startswith(".")]
    if len(video_file_paths) != len(subtitle_file_paths):
        print("The numbers of input videos and subtitles do not match")
        parser.print_usage()
        sys.exit(21)

    output_dir = os.path.abspath(FLAGS.output_directory)
    os.makedirs(output_dir, exist_ok=True)
    video_file_paths = sorted(video_file_paths, key=lambda x: os.path.splitext(os.path.basename(x))[0])
    subtitle_file_paths = sorted(subtitle_file_paths, key=lambda x: os.path.splitext(os.path.basename(x))[0])
    exit_segfail = FLAGS.exit_segfail
    stretch = not FLAGS.stretch_off
    stretch_in_lang = FLAGS.stretch_in_language

    from subaligner.logger import Logger
    Logger.VERBOSE = FLAGS.debug
    Logger.QUIET = FLAGS.quiet
    from subaligner.predictor import Predictor
    from subaligner.translator import Translator
    from subaligner.subtitle import Subtitle
    from subaligner.exception import UnsupportedFormatException
    from subaligner.exception import TerminalException

    predictor = Predictor()
    failures = []
    for index in range(len(video_file_paths)):
        local_video_path = video_file_paths[index]
        local_subtitle_path = subtitle_file_paths[index]
        try:
            if FLAGS.mode == "single":
                aligned_subs, audio_file_path, voice_probabilities, frame_rate = predictor.predict_single_pass(
                    video_file_path=local_video_path,
                    subtitle_file_path=local_subtitle_path,
                    weights_dir=os.path.join(FLAGS.training_output_directory, "models/training/weights")
                )
            else:
                aligned_subs, subs, voice_probabilities, frame_rate = predictor.predict_dual_pass(
                    video_file_path=local_video_path,
                    subtitle_file_path=local_subtitle_path,
                    weights_dir=os.path.join(FLAGS.training_output_directory, "models/training/weights"),
                    stretch=stretch,
                    stretch_in_lang=stretch_in_lang,
                    exit_segfail=exit_segfail,
                )

            parent_dir = os.path.dirname(local_subtitle_path.replace(os.path.abspath(FLAGS.subtitle_directory), output_dir))
            os.makedirs(parent_dir, exist_ok=True)
            aligned_subtitle_path = os.path.abspath(os.path.join(parent_dir,
                                                    ".".join(os.path.basename(local_subtitle_path).rsplit(".", 1)).replace(".stl", ".srt")))

            if FLAGS.translate is not None:
                source, target = FLAGS.translate.split(",")
                translator = Translator(source, target)
                aligned_subs = translator.translate(aligned_subs)
                Subtitle.export_subtitle(local_subtitle_path, aligned_subs, aligned_subtitle_path, frame_rate, "utf-8")
            else:
                Subtitle.export_subtitle(local_subtitle_path, aligned_subs, aligned_subtitle_path, frame_rate)

            log_loss = predictor.get_log_loss(voice_probabilities, aligned_subs)
            if log_loss is None or log_loss > FLAGS.max_logloss:
                print(
                    "Alignment failed with a too high loss value: {} for {} and {}".format(log_loss, local_video_path, local_subtitle_path)
                )
                failures.append((local_video_path, local_subtitle_path))
                continue

            print("Aligned subtitle saved to: {}".format(aligned_subtitle_path))
        except UnsupportedFormatException as e:
            print(
                "{}\n{}".format(str(e), "".join(traceback.format_stack()) if FLAGS.debug else "")
            )
            traceback.print_tb(e.__traceback__)
            failures.append((local_video_path, local_subtitle_path))
            continue
        except TerminalException as e:
            print(
                "{}\n{}".format(str(e), "".join(traceback.format_stack()) if FLAGS.debug else "")
            )
            traceback.print_tb(e.__traceback__)
            failures.append((local_video_path, local_subtitle_path))
            continue
        except Exception as e:
            print(
                "{}\n{}".format(str(e), "".join(traceback.format_stack()) if FLAGS.debug else "")
            )
            traceback.print_tb(e.__traceback__)
            failures.append((local_video_path, local_subtitle_path))
            continue
        else:
            continue

    if len(failures) > 0:
        print("WARNING: The following video and subtitle failed to align with each other:")
        for failure in failures:
            video_file_path, subtitle_file_path = failure
            print("\t{}  {}".format(video_file_path, subtitle_file_path))


if __name__ == "__main__":
    main()
